using System;
using UnityEngine;
using System.Collections;
using System.Collections.Generic;
using System.Linq;
using SimpleJSON;
using MVR.FileManagementSecure;
using UnityEngine.UI;
using Unity.Jobs.LowLevel.Unsafe;
using static UnityEngine.Application;
using System.Security.Policy;


namespace JayJayWon
{
    public class ResourceActionsUI
    {
        private GameObject actionsUIBoxGO;
        private GameObject tabsUIBoxGO;
        private BrowserUI parentBrowser;

        private Dictionary<int, ResourceActionUIBase> resourceActionUITabs = new Dictionary<int, ResourceActionUIBase>();

        private List<UIDynamicButton> resourceActionTabButtons = new List<UIDynamicButton>();

        public int currentActionUITabType = ActionUITabType.primaryActionTab;

        protected int currentResourceType { get { return parentBrowser.browserFiltersUI.currentResourceType; } }

        protected int currentResourceCategory { get { return ResourceType.GetResourceCategory(currentResourceType); } }

        public const float resourceActionTabsHeight = 15f;
        const float resourceActionTabsWidth = 50f;

        public void SetActive(bool active)
        {
            actionsUIBoxGO.SetActive(active);
            tabsUIBoxGO.SetActive(active);
        }

        public ResourceActionUIBase currentActionUITab
        {
            get
            {
                if (resourceActionUITabs.ContainsKey(currentActionUITabType)) return resourceActionUITabs[currentActionUITabType];
                return null;
            }
        }

        public ResourceActionUIBase getCurrentActionUITab
        {
            get
            {
                List<int> actionChoices = GetActionUITabChoices();
                if (!actionChoices.Contains(currentActionUITabType))
                {
                    if (resourceActionUITabs.ContainsKey(currentActionUITabType)) resourceActionUITabs[currentActionUITabType].ClearUI();
                    currentActionUITabType = actionChoices[0];
                    RefreshActionTabs();
                }
                if (resourceActionUITabs.ContainsKey(currentActionUITabType)) return resourceActionUITabs[currentActionUITabType];
                return CreateActionTab(currentActionUITabType);
            }
        }
        public ResourcePrimaryActionUI primaryActionUITab
        {
            get
            {
                List<int> actionChoices = GetActionUITabChoices();

                if (resourceActionUITabs.ContainsKey(ActionUITabType.primaryActionTab)) return (ResourcePrimaryActionUI)resourceActionUITabs[ActionUITabType.primaryActionTab];
                return (ResourcePrimaryActionUI)CreateActionTab(ActionUITabType.primaryActionTab);
            }
        }
        private void RefreshActionTabs()
        {
            var actionTabTypes = GetActionUITabChoices();
            int currentTabButtonCount = resourceActionTabButtons.Count;
            int requiredTabButtonCount = actionTabTypes.Count;

            if (currentTabButtonCount < requiredTabButtonCount)
            {
                for (int i = currentTabButtonCount; i < requiredTabButtonCount; i++) resourceActionTabButtons.Add(CreateResourceActionTabButton(i));
            }
            else if (currentTabButtonCount > requiredTabButtonCount)
            {
                for (int i = currentTabButtonCount; i > requiredTabButtonCount; i--)
                {
                    GameObject.Destroy(resourceActionTabButtons[i-1].gameObject);
                    resourceActionTabButtons.RemoveAt(i-1);
                }
            }
            int index = 0;
            foreach (var tab in resourceActionTabButtons)
            {
                tab.label = ActionUITabType.GetTabLabel(actionTabTypes[index]);
                int actionTabType = actionTabTypes[index];
                if (actionTabType == currentActionUITabType)
                {
                    tab.buttonColor = BrowserUI.darkBlueGrey;
                    tab.textColor = Color.white;
                }
                else
                {
                    tab.buttonColor = BrowserUI.lightGrey;
                    tab.textColor = Color.black;
                }
                var buttonUID = resourceActionTabButtons[index];
                buttonUID.button.onClick.RemoveAllListeners();
                buttonUID.button.onClick.AddListener(delegate {
                    TabSelected(actionTabType);
                });
                AssignTabHelpText(buttonUID, actionTabType);
                index++;
            }
        }

        private void AssignTabHelpText(UIDynamicButton buttonUID, int actionTabType)
        {
            string helpText = "";

            if (actionTabType == ActionUITabType.primaryActionTab) helpText = "Perform primary actions on the selected resource. Where multiple resource are selected, the primary action will be performed on the last selected resource.";
            if (actionTabType == ActionUITabType.tagsTab)
            {
                if (currentResourceCategory == ResourceCategory.clothing) helpText = "Assign tags to ALL selected resources. BA Tags will be synced with VAM Clothing tags on VAM restart. VAM Tag changes can be synced using the Rescan button in BA. Tag changes to clothing in both VAM and BA without syncing may result in conflicts.";
                else helpText = "Assign tags to ALL the selected resources.";
            }
            if (actionTabType == ActionUITabType.positionTab) helpText = "Adjust the position of Scene Atoms";
            if (actionTabType == ActionUITabType.infoTab) helpText = "View and modify info and meta data assoicated with the selected resource(s)";
            if (actionTabType == ActionUITabType.saveTab) helpText = "Save a new resource or overwrite the selected resource (e.g. Preset or Scene).";
            parentBrowser.ApplyHelp(buttonUID, helpText);
        }

        public void RefreshSelectionFocus(int tabType)
        {
            if (tabType == ActionUITabType.tagsTab || (tabType == ActionUITabType.primaryActionTab && currentResourceCategory == ResourceCategory.clothing)) parentBrowser.resourceSelectionsUI.SetAllFocus();
            else parentBrowser.resourceSelectionsUI.SetSingleFocus();
        }

        private void TabSelected(int tabType)
        {
            if (currentActionUITabType == ActionUITabType.tagsTab) (getCurrentActionUITab as ResourceTagsActionUI).CancelTagExpansionMode();

            if (currentActionUITabType != tabType)
            {
                getCurrentActionUITab.ClearUI();
                currentActionUITabType = tabType;
                RefreshUI();
                RefreshSelectionFocus(tabType);
            }
        }
        private UIDynamicButton CreateResourceActionTabButton(int i)
        {
            UIDynamicButton newTabButton = UI.CreateUIButton(tabsUIBoxGO, resourceActionTabsWidth, resourceActionTabsHeight, i.ToString(), 20, i * (resourceActionTabsWidth+1), anchorPreset: AnchorPresets.bottomLeft);
            newTabButton.buttonImage.sprite = null;
            return newTabButton;
        }

        public void RefreshUI()
        {
            RefreshActionTabs();
            getCurrentActionUITab.RefreshUI();
            
        }

        public void RefreshUIResourceSelectionChange()
        {
            if (currentActionUITabType == ActionUITabType.tagsTab) (getCurrentActionUITab as ResourceTagsActionUI).RefreshTagToggles();
            if (currentActionUITabType == ActionUITabType.infoTab) getCurrentActionUITab.RefreshUI();
            if (currentActionUITabType == ActionUITabType.primaryActionTab) getCurrentActionUITab.RefreshUI();
            if (currentActionUITabType == ActionUITabType.saveTab) getCurrentActionUITab.RefreshUI();
        }

        public ResourceActionsUI(BrowserUI _parentBrowser, GameObject _parentBoxGO)
        {
            parentBrowser = _parentBrowser;

            tabsUIBoxGO = UI.AddChildGOImage(parentGO: _parentBoxGO,
                color: Color.grey,
                anchorPreset: AnchorPresets.topMiddle,
                horizontalSize: 640f,
                verticalSize: resourceActionTabsHeight,
                anchoredPositionOffset: new Vector2(0f, 0f));

            actionsUIBoxGO = UI.AddChildGOImage(parentGO: _parentBoxGO,
                color: BrowserUI.darkBlueGrey,
                anchorPreset: AnchorPresets.bottomMiddle,
                horizontalSize: 640f,
                verticalSize: _parentBrowser.resourceActionBoxHeight - resourceActionTabsHeight,
                anchoredPositionOffset: new Vector2(0f, 0f));


            ResourcePositionActionUI.existingAtomPosition = Globals.mvrScript.gameObject.AddComponent<AtomPositionRotationSync>();

            ResourcePositionActionUI.newAtomPosition = Globals.mvrScript.gameObject.AddComponent<RelativeAtomPositionRotation>();

            ResourcePositionActionUI.newAtomPosition.InitNoCallback();
        }

        public void AtomNameUpdated(string oldName, string newName)
        {
            foreach (var kvp in resourceActionUITabs) kvp.Value.AtomNameUpdated(oldName, newName);
        }

        public void OnAtomAdded(Atom atom)
        {
            primaryActionUITab.RefreshTargetAtomPopup();
            if (primaryActionUITab.targetAtomJSMultiEnum.valType==JSONStorableMultiEnumStringChooser.topEnumValFlag) primaryActionUITab.targetAtomJSMultiEnum.mainVal = atom.name;
            if (BrowserAssist.mainBrowserUI.isVisible) primaryActionUITab.RefreshTargetAtomValue();

        }

        public void OnAtomRemoved(Atom atom)
        {
            if (BrowserAssist.mainBrowserUI.isVisible) primaryActionUITab.RefreshTargetAtomValue();

            foreach (var kvp in resourceActionUITabs) kvp.Value.OnAtomRemoved(atom);

        }

        private ResourceActionUIBase CreateActionTab(int actionTabType)
        {
            ResourceActionUIBase newActionTab = null;
            if (actionTabType == ActionUITabType.primaryActionTab) newActionTab = new ResourcePrimaryActionUI(parentBrowser, actionsUIBoxGO);
            if (actionTabType == ActionUITabType.tagsTab) newActionTab = new ResourceTagsActionUI(parentBrowser, actionsUIBoxGO);
            if (actionTabType == ActionUITabType.positionTab) newActionTab = new ResourcePositionActionUI(parentBrowser, actionsUIBoxGO, primaryActionUITab);
            if (actionTabType == ActionUITabType.infoTab) newActionTab = new ResourceInfoActionUI(parentBrowser, actionsUIBoxGO);
            if (actionTabType == ActionUITabType.saveTab) newActionTab = new ResourceSaveActionUI(parentBrowser, actionsUIBoxGO, primaryActionUITab);

            if (newActionTab != null)
            {
                if (resourceActionUITabs.ContainsKey(actionTabType)) resourceActionUITabs[actionTabType] = newActionTab;
                else resourceActionUITabs.Add(actionTabType, newActionTab);
            }
            return (newActionTab);
        }

        private List<int> GetActionUITabChoices()
        {
            List<int> choices = new List<int>();

            choices.Add(ActionUITabType.primaryActionTab);
            

            if (currentResourceType != ResourceType.scene && currentResourceType != ResourceType.plugins) choices.Add(ActionUITabType.positionTab);            
            choices.Add(ActionUITabType.tagsTab);
            if (currentResourceType != ResourceType.plugins && currentResourceType != ResourceType.subScenes && currentResourceCategory != ResourceCategory.clothing && currentResourceType!= ResourceType.cua) choices.Add(ActionUITabType.saveTab);
            choices.Add(ActionUITabType.infoTab);

            return choices;
        }

    }


    public abstract class ResourceActionUIBase
    {
        protected GameObject parentBoxGO;
        public BrowserUI parentBrowser { get; protected set; }

        protected const float edgeWidth = 2f;

        protected HashSet<UIDynamic> uiElements = new HashSet<UIDynamic>();

        protected HashSet<UIDynamicPopup> uiPopups = new HashSet<UIDynamicPopup>();

        protected int resourceType { get { return parentBrowser.browserFiltersUI.currentResourceType; } }

        protected string atomPresetResourceType { get { return parentBrowser.browserFiltersUI.currentAtomPresetResourceType; } }

        protected int resourceCategory { get { return ResourceType.GetResourceCategory(resourceType); } }

        public ResourceActionUIBase(BrowserUI _parentBrowser, GameObject _parentBoxGO)
        {
            parentBrowser = _parentBrowser;
            parentBoxGO = _parentBoxGO;

        }

        public virtual void AtomNameUpdated(string oldName, string newName)
        {
        }
        public virtual void OnAtomRemoved(Atom atom)
        {
        }

        public virtual void RefreshUI()
        {
            ClearUI();
        }

        public virtual void ClearUI()
        {
            foreach (var popup in uiPopups) popup.popup.visible = false;
            foreach (var uiElement in uiElements) if (uiElement!=null) GameObject.Destroy(uiElement.gameObject);
            uiElements.Clear();
        }

        public UIDynamicSlider CreateSlider(GameObject parentGO, JSONStorableFloat jsf, float width, float height, string label = "", int fontSize = 20, float anchorXOffset = 0, float anchorYOffset = 0, int anchorPreset = AnchorPresets.none, float quickButtonHeight = 0f, float labelHeight = 0f, float rangeWidth = 0f, bool hideRange = false, bool hideRangeLabel = false)
        {
            UIDynamicSlider slider = UI.CreateSlider(parentGO, jsf, width, height, label, fontSize, anchorXOffset, anchorYOffset, anchorPreset,quickButtonHeight,labelHeight,rangeWidth, hideRange, hideRangeLabel);
            uiElements.Add(slider);
            return slider;
        }

        public UIDynamicButton CreateUIButton(GameObject parentGO, float width, float height, string label = "", int fontSize = 20, float anchorXOffset = 0, float anchorYOffset = 0, int anchorPreset = AnchorPresets.none, bool clearSprite = false)
        {
            UIDynamicButton button = UI.CreateUIButton(parentGO, width, height, label, fontSize, anchorXOffset, anchorYOffset, anchorPreset, clearSprite);
            uiElements.Add(button);
            return button;
        }

        public UIDynamicPopup CreateScrollablePopup(GameObject parentGO, JSONStorableStringChooser jssc, float width, float height, string label = "", int labelFontSize = 20, float anchorXOffset = 0, float anchorYOffset = 0, float popupHeight = 300f, bool popupAbove = false, bool disableSlider=false, int anchorPreset = AnchorPresets.none, int popupFontSize =20, float labelProportion = 0.4f)
        {
            UIDynamicPopup popup = UI.CreateScrollablePopup(parentGO, jssc, width, height, label, labelFontSize, anchorXOffset, anchorYOffset, popupHeight,popupAbove, disableSlider, anchorPreset, popupFontSize: popupFontSize, labelProportion:labelProportion);
            uiElements.Add(popup);
            uiPopups.Add(popup);
            return popup;
        }

        public UIDynamicPopup CreatePopup(GameObject parentGO, JSONStorableStringChooser jssc, float width, float height, string label = "", int fontSize = 20, float anchorXOffset = 0, float anchorYOffset = 0, int anchorPreset = AnchorPresets.none, bool popAbove = false, float labelProportion=0.4f)
        {
            UIDynamicPopup popup = UI.CreatePopup(parentGO, jssc, width, height, label, fontSize, anchorXOffset, anchorYOffset, anchorPreset, popAbove, labelProportion: labelProportion);
            uiElements.Add(popup);
            uiPopups.Add(popup);
            return popup;
        }

        public UIDynamicTextField CreateTextField(GameObject parentGO, JSONStorableString jss, float width, float height, int fontSize = 20, float anchorXOffset = 0, float anchorYOffset = 0, int anchorPreset = AnchorPresets.none, bool input = false, InputField.LineType inputFieldType = InputField.LineType.MultiLineNewline, string placeHolderText = "", float hScrollBarHeight = 0f, float vScrollBarWidth = 10f, Action<string> textEntryChangeCallback = null)
        {
            UIDynamicTextField textField = UI.CreateTextBox(parentGO, jss, width, height, fontSize, anchorXOffset, anchorYOffset, anchorPreset, input, inputFieldType, placeHolderText, hScrollBarHeight, vScrollBarWidth, textEntryChangeCallback);
            uiElements.Add(textField);
            return textField;
        }

        public UIDynamicToggle CreateToggle(GameObject parentGO, JSONStorableBool jsb, float width, float height, string label = "", int fontSize = 20, float anchorXOffset = 0, float anchorYOffset = 0, int anchorPreset = AnchorPresets.none)
        {
            UIDynamicToggle toggle = UI.CreateToggle(parentGO, jsb, width, height, label, fontSize, anchorXOffset, anchorYOffset, anchorPreset);
            uiElements.Add(toggle);
            return toggle;
        }
        public TriStateToggle CreateTriStateToggle(GameObject parentGO, JSONStorableBool jsb, float width, float height, string label = "", int fontSize = 20, float anchorXOffset = 0, float anchorYOffset = 0, int anchorPreset = AnchorPresets.none)
        {
            UIDynamicToggle toggle = UI.CreateToggle(parentGO, jsb, width, height, label, fontSize, anchorXOffset, anchorYOffset, anchorPreset);
            uiElements.Add(toggle);
            return new TriStateToggle(jsb, toggle);
        }


    }

    public class ResourcePrimaryActionUI : ResourceActionUIBase
    {
        public JSONStorableMultiEnumStringChooser targetAtomJSMultiEnum;

        public CustomUnityAssetLoader cuaLoader { get; private set; } = null;
        private JSONStorableString cuaURLJSS = new JSONStorableString("cuaURL", "");
        private JSONStorableBool cuaImportLightMapsJSB ;
        private JSONStorableBool cuaLoadDllsJSB ;
        private JSONStorableBool cuaRegisterUICanvasesJSB;
        private JSONStorableBool cuaImportLightProbesJSB;
        private JSONStorableBool cuaShowUICanvasesJSB;
        private JSONStorableStringChooser cuaAssetNameJSSC ;

        private UIDynamicTextField cuaURLTextField = null;
        private UIDynamicButton cuaClearButton =null;
        private UIDynamicPopup cuaNamePopup = null;
        private GameObject cuaOptionsPopupGO = null;
        public JSONStorableBool mergeLoadJSB
        {
            get
            {
                var settings = Settings.GetResourceSettings(resourceType, atomPresetResourceType);
                return settings.mergeLoadJSB;
            }
        }
        public JSONStorableBool openPluginUIJSB;
        public JSONStorableBool suppressScaleChangeJSB;

        public JSONStorableBool morphPresetIncludePhysicalJSB;
        public JSONStorableBool morphPresetIncludeAppJSB;

        public JSONStorableBool generalPresetIncludePhysicalJSB;
        public JSONStorableBool generalPresetIncludePoseJSB;
        public JSONStorableBool generalPresetIncludeAppJSB;

        private static Dictionary<Atom, HashSet<DAZClothingItem>> cachedDCIs = new Dictionary<Atom, HashSet<DAZClothingItem>>();

        public bool isAtomTargetable { get
            {
                if (resourceCategory == ResourceCategory.preset) return true;
                if (resourceCategory == ResourceCategory.subScenes) return true;
                if (resourceCategory == ResourceCategory.cua) return true;
                if (resourceCategory == ResourceCategory.clothing) return true;
                if (resourceType == ResourceType.plugins && targetAtomJSMultiEnum.valType != JSONStorableMultiEnumStringChooser.topEnumValFlag) return true;
                return false;
            }
        }

        public bool isCUAResourceType { get { return resourceType== ResourceType.cua; } }

        private int resoureTargetCategory { get
            {
                return ResourceType.GetResourceTargetCategory(resourceType);
            } }

        private ResourceVersionGroupEntry primaryRVGE { get { return parentBrowser.resourceSelectionsUI.GetFirstSelectedRVGE(resourceType, atomPresetResourceType); } }

        private HashSet<ResourceVersionGroupEntry> currentRVGESelections { get
            {
                var selections = new HashSet<ResourceVersionGroupEntry>();

                if (parentBrowser.browserFiltersUI.selectModeJSEnum.val == BrowserSelectMode.multi)
                {
                    selections = new HashSet<ResourceVersionGroupEntry>(parentBrowser.resourceSelectionsUI.GetResourceTypeSelections(resourceType, atomPresetResourceType).selectedRVGEs);
                }
                else if (primaryRVGE != null) selections.Add(primaryRVGE);

                return selections;

            } }

        private UIDynamicPopup targetAtomPopup;
        private UIDynamicButton mainActionButton;
        public bool deleteConfirmationActive { get; set; } = false;
        public bool bulkDeleteConfirmationActive { get; set; } = false;

        public ResourcePrimaryActionUI(BrowserUI _parentBrowser, GameObject _parentBoxGO) : base(_parentBrowser, _parentBoxGO)
        {
            targetAtomJSMultiEnum = new JSONStorableMultiEnumStringChooser("targetAtom", null, null, null, null, TargetAtomMainEnumCallBack, CreateNewAtom.enumManifestName, TargetAtomTopEnumCallBack);
            targetAtomJSMultiEnum.label = "Target Atom";
            RefreshTargetAtomPopup(true);

            openPluginUIJSB = new JSONStorableBool("openPluginUI", false);

            suppressScaleChangeJSB = new JSONStorableBool("suppressClothingLoad", false);

            morphPresetIncludePhysicalJSB = new JSONStorableBool("morphPresetIncludePhysical", false);
            morphPresetIncludeAppJSB = new JSONStorableBool("morphPresetIncludeApp", true);

            generalPresetIncludePhysicalJSB = new JSONStorableBool("generalPresetIncludePhysical", true);
            generalPresetIncludePoseJSB = new JSONStorableBool("generalPresetIncludePose", true);
            generalPresetIncludeAppJSB = new JSONStorableBool("generalPresetIncludeApp", true);

            cuaAssetNameJSSC = new JSONStorableStringChooser("cuaAssetName", null, "", "Asset Name", BACUANameCallback);
            cuaImportLightMapsJSB = new JSONStorableBool("cuaImportLightMaps", true, BACUAImportLighmapsCallback);
            cuaLoadDllsJSB = new JSONStorableBool("cuaLoadDlls", true,BACUALoadDllCallback);
            cuaRegisterUICanvasesJSB = new JSONStorableBool("cuaRegisterUICanvases", true, BACUARegisterUICanvasesCallback);
            cuaImportLightProbesJSB = new JSONStorableBool("cuaImportLightProbes", true,BACUAImportLightProbesCallback);
            cuaShowUICanvasesJSB = new JSONStorableBool("cuaShowUICanvases", true, BACUAShowUICanvasesCallback);
        }

        private void ApplyCUAUrlHelp()
        {
            if (cuaURLTextField!=null && cuaLoader!=null)
            {
                var assetURLJSS = cuaLoader.GetUrlJSONParam("assetUrl");
                parentBrowser.ApplyHelp(cuaURLTextField, "Loaded CUA Bundle file: '" + assetURLJSS.val+"'");
            }
        }
        private void CUAAssetURLCallback(string url)
        {
            string fileName = FileManagerSecure.GetFileName(url);
            if (fileName.Contains(".")) fileName = fileName.Substring(0, fileName.IndexOf("."));
            cuaURLJSS.val = fileName;
            if (cuaURLTextField!= null)
            {
                if (url == "")
                {
                    cuaURLTextField.gameObject.SetActive(false);
                    cuaNamePopup.gameObject.SetActive(false);
                    cuaClearButton.gameObject.SetActive(false);
                }
                else
                {
                    cuaURLTextField.gameObject.SetActive(true);
                    cuaNamePopup.gameObject.SetActive(true);
                    cuaClearButton.gameObject.SetActive(true);
                }
            }
            ApplyCUAUrlHelp();
        }

        private void BACUANameCallback(string assetName)
        {
            if (cuaLoader != null)
            {
                var cuaAssetNameJSSC = cuaLoader.GetStringChooserJSONParam("assetName");
                cuaAssetNameJSSC.val = this.cuaAssetNameJSSC.val;
            }
        }
        private void BACUAImportLighmapsCallback(bool value)
        {
            if (cuaLoader != null)
            {
                var cuaImportLightMapsJSB = cuaLoader.GetBoolJSONParam("importLightmaps");
                cuaImportLightMapsJSB.val = this.cuaImportLightMapsJSB.val;
            }
        }

        private void BACUALoadDllCallback(bool value)
        {
            if (cuaLoader != null)
            {
                var cuaLoadDLLJSB = cuaLoader.GetBoolJSONParam("loadDll");
                cuaLoadDLLJSB.val = this.cuaLoadDllsJSB.val;
            }
        }

        private void BACUAImportLightProbesCallback(bool value)
        {
            if (cuaLoader != null)
            {
                var cuaImportLightProbesJSB = cuaLoader.GetBoolJSONParam("importLightProbes");
                cuaImportLightProbesJSB.val = this.cuaImportLightProbesJSB.val;
            }
        }
        private void BACUARegisterUICanvasesCallback(bool value)
        {
            if (cuaLoader != null)
            {
                var cuaRegisterCanvasesJSB = cuaLoader.GetBoolJSONParam("registerCanvases");
                cuaRegisterCanvasesJSB.val = this.cuaRegisterUICanvasesJSB.val;
            }
        }
        private void BACUAShowUICanvasesCallback(bool value)
        {
            if (cuaLoader != null)
            {
                var cuaShowCanvasesJSB = cuaLoader.GetBoolJSONParam("showCanvases");
                cuaShowCanvasesJSB.val = this.cuaShowUICanvasesJSB.val;
            }
        }

        private void CUAAssetNameCallback(string assetName)
        {
            cuaAssetNameJSSC.valNoCallback = assetName;
        }

        private void CUAImportLighmapsCallback(bool value)
        {
            cuaImportLightMapsJSB.valNoCallback = value;
        }

        private void CUALoadDllCallback(bool value)
        {
            cuaLoadDllsJSB.valNoCallback = value;
        }
        private void CUAImportLightProbesCallback(bool value)
        {
            cuaImportLightProbesJSB.valNoCallback = value;
        }
        private void CUARegisterUICanvasesCallback(bool value)
        {
            cuaRegisterUICanvasesJSB.valNoCallback = value;
        }
        private void CUAShowUICanvasesCallback(bool value)
        {
            cuaShowUICanvasesJSB.valNoCallback = value;
        }

        private void RefreshCUALoader()
        {
            if (cuaLoader != null)
            {
                var assetURLJSS = cuaLoader.GetUrlJSONParam("assetUrl");
                assetURLJSS.setCallbackFunction -= CUAAssetURLCallback;

                var cuaImportLightMapsJSB = cuaLoader.GetBoolJSONParam("importLightmaps");
                cuaImportLightMapsJSB.setCallbackFunction -= CUAImportLighmapsCallback;

                var cuaLoadDLLJSB = cuaLoader.GetBoolJSONParam("loadDll");
                cuaLoadDLLJSB.setCallbackFunction -= CUALoadDllCallback;

                var cuaImportLightProbesJSB = cuaLoader.GetBoolJSONParam("importLightProbes");
                cuaImportLightProbesJSB.setCallbackFunction -= CUAImportLightProbesCallback;

                var cuaRegisterCanvasesJSB = cuaLoader.GetBoolJSONParam("registerCanvases");
                cuaRegisterCanvasesJSB.setCallbackFunction -= CUARegisterUICanvasesCallback;

                var cuaShowCanvasesJSB = cuaLoader.GetBoolJSONParam("showCanvases");
                cuaShowCanvasesJSB.setCallbackFunction -= CUAShowUICanvasesCallback;

                cuaAssetNameJSSC.setCallbackFunction += CUAAssetNameCallback;
            }
            cuaLoader = null;
            if (isCUAResourceType && targetAtomJSMultiEnum.valType == JSONStorableMultiEnumStringChooser.mainValFlag)
            {
                Atom targetAtom = SuperController.singleton.GetAtomByUid(targetAtomJSMultiEnum.mainVal);
                if (targetAtom != null)
                {
                    var rescaleObjTransform = targetAtom.transform.Find("reParentObject/object/rescaleObject");
                    if (rescaleObjTransform != null)
                    {
                        cuaLoader = rescaleObjTransform.GetComponent <CustomUnityAssetLoader>();
                        var assetURLJSS = cuaLoader.GetUrlJSONParam("assetUrl");
                        assetURLJSS.setCallbackFunction += CUAAssetURLCallback;
                        CUAAssetURLCallback(assetURLJSS.val);

                        var cuaAssetNameJSSC = cuaLoader.GetStringChooserJSONParam("assetName");
                        cuaAssetNameJSSC.setCallbackFunction += CUAAssetNameCallback;
                        CUAAssetNameCallback(cuaAssetNameJSSC.val);

                        var cuaImportLightMapsJSB = cuaLoader.GetBoolJSONParam("importLightmaps");
                        cuaImportLightMapsJSB.setCallbackFunction += CUAImportLighmapsCallback;
                        CUAImportLighmapsCallback(cuaImportLightMapsJSB.val);

                        var cuaLoadDLLJSB = cuaLoader.GetBoolJSONParam("loadDll");
                        cuaLoadDLLJSB.setCallbackFunction += CUALoadDllCallback;
                        CUALoadDllCallback(cuaLoadDLLJSB.val);

                        var cuaImportLightProbesJSB = cuaLoader.GetBoolJSONParam("importLightProbes");
                        cuaImportLightProbesJSB.setCallbackFunction += CUAImportLightProbesCallback;
                        CUAImportLightProbesCallback(cuaImportLightProbesJSB.val);

                        var cuaRegisterCanvasesJSB = cuaLoader.GetBoolJSONParam("registerCanvases");
                        cuaRegisterCanvasesJSB.setCallbackFunction += CUARegisterUICanvasesCallback;
                        CUARegisterUICanvasesCallback(cuaRegisterCanvasesJSB.val);

                        var cuaShowCanvasesJSB = cuaLoader.GetBoolJSONParam("showCanvases");
                        cuaShowCanvasesJSB.setCallbackFunction += CUAShowUICanvasesCallback;
                        CUAShowUICanvasesCallback(cuaShowCanvasesJSB.val);

                    }
                }
            }
        }

        private void TargetAtomTopEnumCallBack(int i)
        {

            if (parentBrowser.resourceActionsUI.currentActionUITab == this)
            {
                RefreshUI();
            }
            RefreshActiveClothing();
        }

        
        private void TargetAtomMainEnumCallBack(string i)
        {
            

            if (parentBrowser.resourceActionsUI.currentActionUITab == this)
            {
                RefreshUI();
            }
            RefreshActiveClothing();
        }

        public static void ClearCachedDCIs()
        {
            cachedDCIs.Clear();
        }

        public static HashSet<DAZClothingItem> GetCachedDCIs(Atom atom)
        {
            if (atom == null || atom.type != "Person") return null;

            if (!cachedDCIs.ContainsKey(atom)) cachedDCIs.Add(atom, AtomUtils.GetDAZClothingItems(atom));

            if (cachedDCIs.Count > 1)
            {
                var atomDCIs = cachedDCIs[atom];
                cachedDCIs.Clear();
                cachedDCIs.Add(atom, atomDCIs);
            }


            return cachedDCIs[atom];
        }

        public bool IsClothingActive(ResourceVersionGroupEntry rvge)
        {
            if (rvge.resourceCategory != ResourceCategory.clothing) return false;

            string clothingUID = null;
            string packageName = null;
            if (!rvge.localFileResource)
            {
                PackagedClothingRVGE clothingRVGE = rvge as PackagedClothingRVGE;
                clothingUID = clothingRVGE.clothingUID;
                packageName = clothingRVGE.packageVersionUID;
            }
            else
            {
                LocalClothingRVGE clothingRVGE = rvge as LocalClothingRVGE;
                clothingUID = clothingRVGE.clothingUID;
            }
            Atom atom = SuperController.singleton.GetAtomByUid(targetAtomJSMultiEnum.mainVal);
            if (atom != null && atom.type == "Person")
            {
                var dciList = GetCachedDCIs(atom);
                foreach (var dci in dciList)
                {
                    if (clothingUID == dci.internalUid && (rvge.localFileResource ||packageName== dci.packageUid)) return true;
                }
            }

            return false;

        }

        public void ActivateClothingUI(ResourceVersionGroupEntry rvge)
        {
            if (rvge.resourceCategory != ResourceCategory.clothing) return;
            Atom atom = SuperController.singleton.GetAtomByUid(targetAtomJSMultiEnum.mainVal);
            if (atom != null && atom.type == "Person")
            {
                string clothingUID = null;
                string packageName = null;

                if (!rvge.localFileResource)
                {
                    PackagedClothingRVGE clothingRVGE = rvge as PackagedClothingRVGE;
                    clothingUID = clothingRVGE.clothingUID;
                    packageName= clothingRVGE.packageVersionUID;
                }
                else
                {
                    LocalClothingRVGE clothingRVGE = rvge as LocalClothingRVGE;
                    clothingUID = clothingRVGE.clothingUID;                   
                }

                var dciList = GetCachedDCIs(atom).Where(dci => dci.internalUid == clothingUID && (rvge.localFileResource || packageName == dci.packageUid));
                if (dciList != null && dciList.Count() > 0) dciList.First().OpenUI();

            }
        }


        public void RefreshActiveClothing()
        {
            if (resourceCategory == ResourceCategory.clothing)
            {
                if (parentBrowser.resourceBrowserUI != null)
                {
                    parentBrowser.resourceBrowserUI. ClearActiveResourceIcons();
                    if (targetAtomJSMultiEnum.valType == JSONStorableMultiEnumStringChooser.mainValFlag)
                    {
                        Atom atom = SuperController.singleton.GetAtomByUid(targetAtomJSMultiEnum.mainVal);
                        if (atom != null && atom.type == "Person")
                        {
                            var dcis = GetCachedDCIs(atom);

                            var activeClothingRVGEs = ResourceManifest.GetMatchingClothingRVGEs(dcis, resourceType);
                            foreach (var rvge in activeClothingRVGEs) parentBrowser.resourceBrowserUI.SetActiveResourceIcon(rvge);
                        }
                    }

                }
                UpdateMainButtonLabelClothing();
            }

        }

        public override void AtomNameUpdated(string oldName, string newName)
        {
            if (targetAtomJSMultiEnum.valType != JSONStorableMultiEnumStringChooser.topEnumValFlag)
            {
                if (targetAtomJSMultiEnum.mainVal == oldName)
                {
                    ResetTargetAtomChoices(GetTargetAtomNames());
                    targetAtomJSMultiEnum.mainVal = newName;
                }
            }
        }

        public override void OnAtomRemoved(Atom atom)
        {
            if (targetAtomJSMultiEnum.valType != JSONStorableMultiEnumStringChooser.topEnumValFlag)
            {
                if (targetAtomJSMultiEnum.mainVal == atom.name) RefreshTargetAtomValue();
            }
        }

        private bool createNewAtom { get { return targetAtomJSMultiEnum.valType == JSONStorableMultiEnumStringChooser.topEnumValFlag; } }

        private string targetAtomName { get
            {
                if (!isAtomTargetable) return null;
                if (createNewAtom) return AtomUtils.GetNextAvailableAtomName(atomPresetResourceType);
                return targetAtomJSMultiEnum.mainVal;
            } }
        public void ClothingActivate()
        {
            HashSet<ResourceVersionGroupEntry> clothingToToggle = null;
            var currentSelections = currentRVGESelections;

            var activeClothing = new HashSet<ResourceVersionGroupEntry>();
            if (createNewAtom) clothingToToggle = currentSelections;
            else
            {
                Atom atom = SuperController.singleton.GetAtomByUid(targetAtomJSMultiEnum.mainVal);
                if (atom != null)
                {
                    foreach (var rvge in currentSelections)
                    {
                        var activeJSB = AtomUtils.GetClothingActive(atom, rvge);
                        if (activeJSB != null && activeJSB.val) activeClothing.Add(rvge);
                    }
                }
                else return;

                if (activeClothing.Count == currentSelections.Count) clothingToToggle = activeClothing;
                else clothingToToggle = new HashSet<ResourceVersionGroupEntry>( currentSelections.Except (activeClothing));
            }

            if (clothingToToggle != null && clothingToToggle.Count>0)
            {
                if (createNewAtom)
                {
                    Globals.mvrScript.StartCoroutine(clothingToToggle.First().Activate(parentBrowser, this, createNewAtom, targetAtomName, delegate (string newAtomName)
                    {
                        foreach (var rvge in clothingToToggle.Skip(1))
                        {
                            Globals.mvrScript.StartCoroutine(rvge.Activate(parentBrowser, this, false, newAtomName));
                        }
                    }));
                }
                else foreach (var rvge in clothingToToggle) Activate(rvge);
            }         
        }

        public void Activate(ResourceVersionGroupEntry rvge)
        {
            Globals.mvrScript.StartCoroutine(rvge.Activate(parentBrowser, this, createNewAtom, targetAtomName));
        }

        public void UpdateMainButtonLabelClothing()
        {
            if (targetAtomJSMultiEnum.valType == JSONStorableMultiEnumStringChooser.topEnumValFlag)
            {
                mainActionButton.label = "Wear Selected Clothing";
                return;
            }
            else
            {
                Atom atom = SuperController.singleton.GetAtomByUid(targetAtomJSMultiEnum.mainVal);
                if (atom != null)
                {
                    bool notEnabledClothing = false;
                    foreach (var rvge in currentRVGESelections)
                    {
                        var activeJSB = AtomUtils.GetClothingActive(atom, rvge);
                        if (activeJSB == null || !activeJSB.val)
                        {
                            notEnabledClothing = true;
                            break;
                        }
                    }
                    if (notEnabledClothing)
                    {
                        mainActionButton.label = "Wear Selected Clothing";
                        return;
                    }
                }
                else
                {
                    mainActionButton.label = "Wear Selected Clothing";
                    return;
                }
            }
            mainActionButton.label = "Remove Selected Clothing";

        }
        public override void ClearUI()
        {
            base.ClearUI();
            if (cuaOptionsPopupGO != null) GameObject.Destroy(cuaOptionsPopupGO);
        }
        public override void RefreshUI()
        {
            base.RefreshUI();

            RefreshCUALoader();

            //Merge Load Toggle
            if (resourceType==ResourceType.scene || (resourceCategory == ResourceCategory.preset && targetAtomJSMultiEnum.valType== JSONStorableMultiEnumStringChooser.mainValFlag))
            {
                var settings = Settings.GetResourceSettings(resourceType, atomPresetResourceType);
                var mergeToggle = CreateToggle(parentBoxGO, mergeLoadJSB, 80f, 20f, "Merge Load", 20, edgeWidth, -edgeWidth,AnchorPresets.topLeft);
                parentBrowser.ApplyHelp(mergeToggle, "Enable Merge Load option when loading the resource.");
                mergeToggle.toggle.onValueChanged.AddListener(delegate
                {
                    if (resourceType == ResourceType.presetAppearance || resourceType == ResourceType.presetClothing) RefreshUI();
                });
            }

            // Target Atom
            if (resourceCategory == ResourceCategory.preset || resourceType == ResourceType.plugins || resourceType == ResourceType.subScenes || resourceCategory == ResourceCategory.clothing || resourceType == ResourceType.cua)
            {
                targetAtomJSMultiEnum.setTopEnumCallbackFunction = TargetAtomTopEnumCallBack;
                targetAtomJSMultiEnum.setMainValCallbackFunction = TargetAtomMainEnumCallBack;

                targetAtomPopup = CreateScrollablePopup(parentBoxGO, targetAtomJSMultiEnum, 160f, 30f, "Target Atom", 20, edgeWidth, -edgeWidth - 20f, 300f, true, true, AnchorPresets.topLeft);
                targetAtomPopup.popup.onOpenPopupHandlers += RefreshTargetAtomPopupCallback;
                parentBrowser.ApplyHelp(targetAtomPopup, "Select a target atom from the current scene that the primary action for the selected resource will be performed on.");
                RefreshTargetAtomValue();
            }
            if (cuaLoader != null)
            {

                cuaURLTextField = CreateTextField(parentBoxGO, cuaURLJSS, 160f, 20f, 20, edgeWidth, -edgeWidth - 50f, AnchorPresets.topLeft,vScrollBarWidth:0f);
                ApplyCUAUrlHelp();

                cuaClearButton = CreateUIButton(parentBoxGO, 40f, 20f, "Clear", 20, edgeWidth + 160f, -edgeWidth - 50f, AnchorPresets.topLeft);
                cuaClearButton.buttonColor = Color.red;
                cuaClearButton.buttonText.color = Color.white;
                parentBrowser.ApplyHelp(cuaClearButton, "Clear loaded asset bundle from CUA atom '"+targetAtomName+"'");
                cuaClearButton.button.onClick.AddListener(delegate
                {
                    if (cuaLoader != null)
                    {
                        var assetURLJSS = cuaLoader.GetUrlJSONParam("assetUrl");
                        assetURLJSS.val = "";

                        RefreshUI();
                    }
                        
                });


                cuaNamePopup = CreateScrollablePopup(parentBoxGO,cuaAssetNameJSSC,300f,20f,"Asset Name",20,edgeWidth+200f,-edgeWidth-50f,150,true,true,AnchorPresets.topLeft,20,0.25f);
                parentBrowser.ApplyHelp(cuaNamePopup, "Select Asset Name from the bundle loaded on CUA atom '" + targetAtomName + "'");

                cuaNamePopup.popup.onOpenPopupHandlers += delegate
                {
                    if (cuaLoader!=null)
                    {
                        var cuaAssetNameJSSC = cuaLoader.GetStringChooserJSONParam("assetName");
                        this.cuaAssetNameJSSC.choices = cuaAssetNameJSSC.choices;
                        this.cuaAssetNameJSSC.displayChoices = cuaAssetNameJSSC.displayChoices;
                    }
                };

                if (cuaURLJSS.val == "")
                {
                    cuaURLTextField.gameObject.SetActive(false);
                    cuaNamePopup.gameObject.SetActive(false);
                    cuaClearButton.gameObject.SetActive(false);
                }
            }

            // Open Plugins UI Toggle
            if (resourceType == ResourceType.plugins)
            {
                var openPluginUIToggle = CreateToggle(parentBoxGO, openPluginUIJSB, 100f, 20f, "Open Plugin UI", 20, edgeWidth, -edgeWidth, AnchorPresets.topLeft);
                parentBrowser.ApplyHelp(openPluginUIToggle, "The Plugin's UI will be opened in the main VAM UI when the plugin is loaded into the target Atom.");
            }

            // Main Load Button
            string resourceTypeName = parentBrowser.browserFiltersUI.currentResourceTypeName;
            if (resourceCategory == ResourceCategory.preset )
            {
                if (resourceType!=ResourceType.presetAtom) resourceTypeName = resourceTypeName.Substring(7);
                mainActionButton = CreateUIButton(parentBoxGO, 100f, 30f, "Load " + resourceTypeName + " Preset", 20, 165f+ edgeWidth, -edgeWidth-20f, AnchorPresets.topLeft);
            }
            else if (resourceType == ResourceType.scene)
            {
                mainActionButton = CreateUIButton(parentBoxGO, 100f, 30f, "Load Scene", 20, edgeWidth, -edgeWidth-20f, AnchorPresets.topLeft);
            }
            else if (resourceType == ResourceType.subScenes)
            {
                mainActionButton = CreateUIButton(parentBoxGO, 100f, 30f, "Load SubScene", 20, 165f + edgeWidth, -edgeWidth - 20f, AnchorPresets.topLeft);
            }
            else if (resourceType == ResourceType.plugins)
            {
                mainActionButton = CreateUIButton(parentBoxGO, 100f, 30f, "Load Plugin", 20, 165f + edgeWidth, -edgeWidth-20f, AnchorPresets.topLeft);
            }
            else if (resourceType == ResourceType.cua)
            {
                mainActionButton = CreateUIButton(parentBoxGO, 100f, 30f, "Load CUA", 20, 165f + edgeWidth, -edgeWidth - 20f, AnchorPresets.topLeft);
            }
            else if (resourceCategory == ResourceCategory.clothing)
            {
                mainActionButton = CreateUIButton(parentBoxGO, 100f, 30f, "Wear Selected Clothing", 20, 165f + edgeWidth, -edgeWidth - 20f, AnchorPresets.topLeft);
                UpdateMainButtonLabelClothing();
            }
            if (mainActionButton != null)
            {
                mainActionButton.buttonColor = new Color(0, 0.6f, 0);
                mainActionButton.textColor = Color.white;
                mainActionButton.button.onClick.AddListener(delegate
                {
                    if (resourceCategory == ResourceCategory.clothing) ClothingActivate();       
                    else if (primaryRVGE != null) Activate(primaryRVGE);
                    
                    deleteConfirmationActive = false;
                    RefreshUI();
                });
                PrimaryActionPointerEnterExitHelp primaryActionContextHelp = mainActionButton.button.gameObject.AddComponent<PrimaryActionPointerEnterExitHelp>();
                primaryActionContextHelp.Init(parentBrowser.helpTextJSS, parentBrowser.resourceSelectionsUI.GetResourceTypeSelections(resourceType, atomPresetResourceType), this);
                if (resourceType!=ResourceType.scene || !parentBrowser.isVREnabled)
                {
                    var settings = Settings.GetResourceSettings(resourceType, atomPresetResourceType);
                    var closeBrowserUIonLoad = CreateToggle(parentBoxGO, settings.closeBrowserUIonResourceLoadJSB, 140f, 20f, "Close UI on Load", 20, 0, 0, AnchorPresets.topRight);
                    parentBrowser.ApplyHelp(closeBrowserUIonLoad, "Close the BrowserAssist window on Resource Load");
                }
                if (isAtomTargetable)
                {
                    var selectAtomOnLoad = CreateToggle(parentBoxGO, Settings.selectAtomOnLoadJSB, 140f, 20f, "VAM Select Atom on Load", 20, 0, -20f, AnchorPresets.topRight);
                    parentBrowser.ApplyHelp(selectAtomOnLoad, "VAM Select the Target atom on Resource Load");

                }
            }
            if ( resourceCategory == ResourceCategory.clothing && targetAtomJSMultiEnum.valType==JSONStorableMultiEnumStringChooser.mainValFlag)
            {

                string label = "Remove All Clothing";
#if VAM_GT_1_21
label = "Remove All Real Clothing";
#endif

                var removeClothingButton = CreateUIButton(parentBoxGO, 120f, 20f, label, 20, edgeWidth, -edgeWidth - 50f, AnchorPresets.topLeft);
                removeClothingButton.button.onClick.AddListener(delegate
                {
                    var targetAtom = SuperController.singleton.GetAtomByUid(targetAtomName);

                    if (targetAtom!= null)
                    {
                        JSONStorable receiver = targetAtom.GetStorableByID("geometry");

                        var dcis = AtomUtils.GetDAZClothingItems(targetAtom);

                        foreach (var dci in dcis)
                        {
                            bool dciRealItem = true;
#if VAM_GT_1_21
dciRealItem = dci.isRealItem;
#endif
                            if (dci.isRealItem)
                            {
                                JSONStorableBool active = receiver.GetBoolJSONParam("clothing:" + dci.uid);
                                if (active != null) active.val = false;
                            }
                        }
                    }

                });
                removeClothingButton.buttonColor = Color.yellow;
                parentBrowser.ApplyHelp(removeClothingButton, "Removes all clothing from the Target Person");
            }

            // Random Load buttons
            if (PC.patreonContentEnabled && resourceCategory!=ResourceCategory.clothing)
            {
                float randomLoadButtonXOffset = 105f + edgeWidth;
                if (resourceCategory == ResourceCategory.preset || resourceType == ResourceType.plugins || resourceType == ResourceType.plugins || resourceType == ResourceType.subScenes || resourceType == ResourceType.cua)
                {
                    randomLoadButtonXOffset = randomLoadButtonXOffset + 165f;
                }
                var loadRandomButton = CreateUIButton(parentBoxGO, 100f, 30f, "Load Random from current Filter results", 20, randomLoadButtonXOffset, -edgeWidth - 20f, AnchorPresets.topLeft);
                loadRandomButton.buttonColor = new Color(0, 0.6f, 0);
                loadRandomButton.textColor = Color.white;
                loadRandomButton.button.onClick.AddListener(delegate
                {
                    var rvge = PC.LoadRandomFromFilterResults(parentBrowser, this);
                    deleteConfirmationActive = false;

                    parentBrowser.resourceSelectionsUI.ForcePrimarySelection(rvge, parentBrowser.resourceBrowserUI.GetUIDButton(rvge), parentBrowser.browserFiltersUI.selectModeJSEnum.val);
                    RefreshUI();
                });
                parentBrowser.ApplyHelp(loadRandomButton, "Load a random resource from all the resources in the browser results");
                if (parentBrowser.resourceSelectionsUI.SelectionCount(resourceType, atomPresetResourceType) > 1 && parentBrowser.browserFiltersUI.selectModeJSEnum.val == BrowserSelectMode.multi)
                {
                    var loadRandomSelectionButton = CreateUIButton(parentBoxGO, 100f, 30f, "Load Random from current Selections", 20, randomLoadButtonXOffset + 105f, -edgeWidth - 20f, AnchorPresets.topLeft);
                    loadRandomSelectionButton.buttonColor = new Color(0, 0.6f, 0);
                    loadRandomSelectionButton.textColor = Color.white;
                    loadRandomSelectionButton.button.onClick.AddListener(delegate
                    {
                        var rvge = PC.LoadRandomFromSelections(parentBrowser, resourceType, atomPresetResourceType, this);
                        deleteConfirmationActive = false;

                        parentBrowser.resourceSelectionsUI.ForcePrimarySelection(rvge, parentBrowser.resourceBrowserUI.GetUIDButton(rvge), parentBrowser.browserFiltersUI.selectModeJSEnum.val);

                        RefreshUI();
                        
                    });
                    SetAllSelectionsEnter allSelectionsEnterExit = loadRandomSelectionButton.button.gameObject.AddComponent<SetAllSelectionsEnter>();
                    allSelectionsEnterExit.Init(parentBrowser.resourceSelectionsUI);
                    parentBrowser.ApplyHelp(loadRandomSelectionButton, "Load a random resource from the resources that are currently selected");
                }
            }
            if (resourceType == ResourceType.cua && targetAtomJSMultiEnum.valType != JSONStorableMultiEnumStringChooser.topEnumValFlag)
            {
                float horizOffset = 265f;
                if (PC.patreonContentEnabled && resourceCategory != ResourceCategory.clothing)
                {
                    horizOffset = 370f;
                    if (parentBrowser.resourceSelectionsUI.SelectionCount(resourceType, atomPresetResourceType) > 1 && parentBrowser.browserFiltersUI.selectModeJSEnum.val == BrowserSelectMode.multi) horizOffset = 475f;
                }

                var cuaLoadOptions = CreateUIButton(parentBoxGO, 20f, 30f, "^", 26, horizOffset + edgeWidth, -edgeWidth - 20f, AnchorPresets.topLeft);
                cuaLoadOptions.button.onClick.AddListener(delegate
                {
                    ToggleCUAOptionPopup();
                });
                parentBrowser.ApplyHelp(cuaLoadOptions, "Show/Hide CUA Load Options");
                CreateCUAOptionPopup(horizOffset);
            }


            // App Preset Suppressions Suppress & Other preset options
            if (resourceType == ResourceType.presetAppearance )
            {                
                var suppressClothingLoadToggle = CreateToggle(parentBoxGO, Settings.suppressClothingLoadJSB, 140f, 20f, "Suppress Clothing Load", 20, edgeWidth, -edgeWidth - 50, AnchorPresets.topLeft);
                parentBrowser.ApplyHelp(suppressClothingLoadToggle, "Any clothing items in the Appearance Preset will be excluded and any clothing on the target atom will reamin.");
               suppressClothingLoadToggle.toggle.onValueChanged.AddListener( delegate
                {
                    RefreshUI();
                });


                float suppressScaleChangeHOffset = 140f;
                if (Settings.suppressClothingLoadJSB.val && (!mergeLoadJSB.val || targetAtomJSMultiEnum.valType == JSONStorableMultiEnumStringChooser.topEnumValFlag))
                {
                    if (Globals.isRealClothingAvailable)
                    {
                        var suppressRealClothingToggle = CreateToggle(parentBoxGO, Settings.onlySuppressRealClothingJSB, 160f, 20f, "Only suppress 'Real' Clothing", 20, edgeWidth + 140f, -edgeWidth - 50, AnchorPresets.topLeft);
                        parentBrowser.ApplyHelp(suppressRealClothingToggle, "Only suppress clothing items in the Appearance Preset that are marked as 'Real'. This can be used to exclude clothing items except those that are 'non-real' such as eye shading effects that are available as clothing items. Keep this setting On if you are unsure.");
                        suppressScaleChangeHOffset = 300f;
                    }
                        
                }

                if (PC.patreonContentEnabled)
                {
                    var suppressScaleChangeToggle = CreateToggle(parentBoxGO, suppressScaleChangeJSB, 140f, 20f, "Suppress Scale Change", 20, edgeWidth + suppressScaleChangeHOffset, -edgeWidth - 50, AnchorPresets.topLeft);
                    parentBrowser.ApplyHelp(suppressScaleChangeToggle, "The scale of the target atom will remain unaffected by the Appearance Preset Load.");
                }
            }
            else if (resourceType == ResourceType.presetClothing)
            {
                if (!mergeLoadJSB.val || targetAtomJSMultiEnum.valType == JSONStorableMultiEnumStringChooser.topEnumValFlag)
                {
                    if (Globals.isRealClothingAvailable)
                    {
                        var onlyReplaceRealClothingToggle = CreateToggle(parentBoxGO, Settings.onlyReplaceRealClothingJSB, 160f, 20f, "Only replace 'Real' Clothing", 20, edgeWidth, -edgeWidth - 50, AnchorPresets.topLeft);
                        parentBrowser.ApplyHelp(onlyReplaceRealClothingToggle, "Only replace clothing items on the Target Atom that are marked as 'Real'. This can be used to keep clothing items that are 'non-real' such as eye shading effects from being modified by the clothing preset load. Keep this setting On if you are unsure.");
                    }
                    
                }
            }
            else if(resourceType == ResourceType.presetMorphs)
            {
                var morphPresetIncludeAppToggle = CreateToggle(parentBoxGO, morphPresetIncludeAppJSB, 120f, 20f, "Include Appearance", 20, edgeWidth, -edgeWidth - 50, AnchorPresets.topLeft);
                parentBrowser.ApplyHelp(morphPresetIncludeAppToggle, "Appearance Morphs will be included in the Preset Load.");
                var morphPresetIncludePhysical = CreateToggle(parentBoxGO, morphPresetIncludePhysicalJSB, 140f, 20f, "Include Physical (Pose)", 20, edgeWidth + 120f, -edgeWidth - 50, AnchorPresets.topLeft);
                parentBrowser.ApplyHelp(morphPresetIncludePhysical, "Pose Morphs will be included in the Preset Load.");
            }
            else if (resourceType == ResourceType.presetGeneral)
            {
                var generalPresetIncludePhysicalToggle = CreateToggle(parentBoxGO, generalPresetIncludePhysicalJSB, 120f, 20f, "Include Physical", 20, edgeWidth, -edgeWidth - 50, AnchorPresets.topLeft);
                parentBrowser.ApplyHelp(generalPresetIncludePhysicalToggle, "Physical Parameters will be included in the General Preset Load.");
                var generalPresetIncludePoseToggle = CreateToggle(parentBoxGO, generalPresetIncludePoseJSB, 100f, 20f, "Include Pose", 20, edgeWidth+120f, -edgeWidth - 50, AnchorPresets.topLeft);
                parentBrowser.ApplyHelp(generalPresetIncludePhysicalToggle, "Pose Parameters will be included in the General Preset Load.");
                var generalPresetIncludeAppToggle = CreateToggle(parentBoxGO, generalPresetIncludeAppJSB, 125f, 20f, "Include Appearance", 20, edgeWidth+220f, -edgeWidth - 50, AnchorPresets.topLeft);
                parentBrowser.ApplyHelp(generalPresetIncludePhysicalToggle, "Appearance Parameters will be included in the General Preset Load.");
            }

            //Delete buttons
            if (primaryRVGE != null && primaryRVGE.isDeletable)
            {
                if (!deleteConfirmationActive && !bulkDeleteConfirmationActive)
                {
                    string label = "Delete\nFile...";
                    string helpText = "Delete '" + primaryRVGE.resourceFullPathReference + "'";
                    if (!primaryRVGE.localFileResource)
                    {
                        label = "Delete\nVAR...";
                        helpText = "Delete the VAR Package containing '" + primaryRVGE.resourceFullPathReference + "'.\nWARNING: This will permanently delete all resources from the package.";
                    }
                    var deleteRVGEButton = CreateUIButton(parentBoxGO, 70f, 30f, label, 22, 0, edgeWidth, AnchorPresets.bottomRight);
                    deleteRVGEButton.buttonColor = Color.red;
                    deleteRVGEButton.buttonText.color = Color.white;
                    if (resourceCategory == ResourceCategory.clothing)
                    {
                        SetSingleSelectionsEnter singleSelectionsEnterExit = deleteRVGEButton.button.gameObject.AddComponent<SetSingleSelectionsEnter>();
                        singleSelectionsEnterExit.Init(parentBrowser.resourceSelectionsUI);
                    }

                    deleteRVGEButton.button.onClick.AddListener(delegate
                    {
                        deleteConfirmationActive = true;
                        RefreshUI();
                    });

                    parentBrowser.ApplyHelp(deleteRVGEButton, helpText, true);

                    if (Settings.enableBulkDeleteJSB.val && parentBrowser.resourceSelectionsUI.SelectionCount(resourceType, atomPresetResourceType) > 1 && parentBrowser.browserFiltersUI.selectModeJSEnum.val == BrowserSelectMode.multi)
                    {
                        var bulkDeleteRVGEButton = CreateUIButton(parentBoxGO, 70f, 30f, "Delete All...", 22, -70f, edgeWidth, AnchorPresets.bottomRight);
                        bulkDeleteRVGEButton.buttonColor = Color.red;
                        bulkDeleteRVGEButton.buttonText.color = Color.white;
                        
                        bulkDeleteRVGEButton.button.onClick.AddListener(delegate
                        {
                            bulkDeleteConfirmationActive = true;
                            RefreshUI();
                        });

                        if (resourceCategory != ResourceCategory.clothing)
                        {
                            SetAllSelectionsEnter allSelectionsEnterExit = bulkDeleteRVGEButton.button.gameObject.AddComponent<SetAllSelectionsEnter>();
                            allSelectionsEnterExit.Init(parentBrowser.resourceSelectionsUI);
                        }
                        

                        parentBrowser.ApplyHelp(bulkDeleteRVGEButton, "Delete all selected resources and any VAR packages containing the selected resources.\nWARNING: This will permanently delete all resources from the packages.", true);
                    }
                }
                else
                {
                    string confirmHelpText = "Confirm deletion of '" + primaryRVGE.resourceFullPathReference + "'";
                    string cancelHelpText = "Cancel deletion of '" + primaryRVGE.resourceFullPathReference + "'";

                    if (bulkDeleteConfirmationActive)
                    {
                        confirmHelpText = "Confirm deletion of ALL selected resources.\nWARNING: This will delete ALL VAR packages containing these resources.";
                        cancelHelpText = "Cancel deletion of selected resources";
                    }

                    if (!bulkDeleteConfirmationActive && !primaryRVGE.localFileResource)
                    {
                        confirmHelpText = "Confirm deletion of VAR containing '" + primaryRVGE.resourceFullPathReference + "'.\nWARNING: This will delete all resources in the containing package.";
                    }
                    var confirmDeleteRVGEButton = CreateUIButton(parentBoxGO, 70f, 30f, "Confirm Deletion", 22, -75f - edgeWidth, edgeWidth, AnchorPresets.bottomRight);
                    confirmDeleteRVGEButton.buttonColor = Color.red;
                    confirmDeleteRVGEButton.buttonText.color = Color.white;

                    if (deleteConfirmationActive)
                    {
                        confirmDeleteRVGEButton.button.onClick.AddListener(delegate
                        {
                            DeletePrimaryRVGE();
                            deleteConfirmationActive = false;
                            RefreshUI();
                        });
                    }
                    else if (bulkDeleteConfirmationActive)
                    {
                        confirmDeleteRVGEButton.button.onClick.AddListener(delegate
                        {
                            DeleteAllSelectedRVGEs();
                            bulkDeleteConfirmationActive = false;
                            RefreshUI();
                        });
                    }
                    
                    parentBrowser.ApplyHelp(confirmDeleteRVGEButton, confirmHelpText, true);

                    var cancelDeleteRVGEButton = CreateUIButton(parentBoxGO, 70f, 30f, "Cancel Deletion", 22, -edgeWidth, edgeWidth, AnchorPresets.bottomRight);
                    cancelDeleteRVGEButton.buttonColor = new Color(0f, 0.6f, 0f);
                    cancelDeleteRVGEButton.buttonText.color = Color.white;
                    cancelDeleteRVGEButton.button.onClick.AddListener(delegate
                    {
                        deleteConfirmationActive = false;
                        bulkDeleteConfirmationActive = false;
                        RefreshUI();

                    });
                    parentBrowser.ApplyHelp(cancelDeleteRVGEButton, cancelHelpText);
                }
                
            }
        }

        private void CreateCUAOptionPopup(float horizOffset)
        {
            cuaOptionsPopupGO = UI.AddChildGOImage(parentBoxGO, Color.gray, AnchorPresets.bottomLeft, 125, 105, new Vector2(horizOffset-85f, 60));

            var importLightmapsToggle = UI.CreateToggle(cuaOptionsPopupGO, cuaImportLightMapsJSB, 120f, 20f, "Import Lightmaps",20,0,-2.5f,AnchorPresets.topMiddle);

            parentBrowser.ApplyHelp(importLightmapsToggle, "Set Import Lightmaps option for the currently selected CUA Atom");

            var importLightProbesToggle = UI.CreateToggle(cuaOptionsPopupGO, cuaImportLightProbesJSB, 120f, 20f, "Import LightProbes", 20, 0, -2.5f-20f, AnchorPresets.topMiddle);

            parentBrowser.ApplyHelp(importLightProbesToggle, "Set Import LightProbes option for the currently selected CUA Atom");

            var loadDllToggle = UI.CreateToggle(cuaOptionsPopupGO, cuaLoadDllsJSB, 120f, 20f, "Load DLL", 20, 0, -2.5f - 40f, AnchorPresets.topMiddle);

            parentBrowser.ApplyHelp(loadDllToggle, "Set Load DLL option for the currently selected CUA Atom");

            var registerCanvasesToggle = UI.CreateToggle(cuaOptionsPopupGO, cuaRegisterUICanvasesJSB, 120f, 20f, "Register UI Canvases", 20, 0, -2.5f - 60f, AnchorPresets.topMiddle);

            parentBrowser.ApplyHelp(registerCanvasesToggle, "Set Register UI Canvases option for the currently selected CUA Atom");

            var showCanvasesToggle = UI.CreateToggle(cuaOptionsPopupGO, cuaShowUICanvasesJSB, 120f, 20f, "Show UI Canvases", 20, 0, -2.5f - 80f, AnchorPresets.topMiddle);

            parentBrowser.ApplyHelp(showCanvasesToggle, "Set Show UI Canvases option for the currently selected CUA Atom");

            cuaOptionsPopupGO.SetActive(false);
        }

        private void ToggleCUAOptionPopup()
        {
            if (cuaOptionsPopupGO.activeSelf) cuaOptionsPopupGO.SetActive(false);
            else
            {
                cuaOptionsPopupGO.SetActive(true);
            }
        }

        private void DeleteRVGE(ResourceVersionGroupEntry rvge)
        {
            bool lastPackagedVersion = false;
            if (rvge.localFileResource)
            {
                FileManagerSecure.DeleteFile(Const.SymLinksFolder + "\\" + rvge.resourceFullPathReference);
                ResourceManifest.PurgeLocalRVGE(rvge as LocalRVGE);

                string thumbnailPath = Const.SymLinksFolder + "\\" + rvge.resourceFolderPathFileNameNoExt + ".jpg";
                string favPath = Const.SymLinksFolder + "\\" + rvge.resourceFolderPathFileNameNoExt + ".fav";
                string hidePath = Const.SymLinksFolder + "\\" + rvge.resourceFolderPathFileNameNoExt + ".hide";
                if (FileManagerSecure.FileExists(thumbnailPath)) FileManagerSecure.DeleteFile(thumbnailPath);
                if (FileManagerSecure.FileExists(favPath)) FileManagerSecure.DeleteFile(favPath);
                if (FileManagerSecure.FileExists(hidePath)) FileManagerSecure.DeleteFile(hidePath);
            }
            else
            {
                PackagedRVGE packagedRVGE = rvge as PackagedRVGE;
                string varFilePath = VARPackageManifest.GetVARPath(packagedRVGE, packagedRVGE.highestVARVersion);

                if (FileManagerSecure.FileExists(Const.SymLinksFolder + "\\" + varFilePath)) FileManagerSecure.DeleteFile(Const.SymLinksFolder + "\\" + varFilePath);

                VARPackageManifest.PurgeVARFile(packagedRVGE.creatorName, packagedRVGE.packageName, packagedRVGE.highestVARVersion);
                if (!VARPackageManifest.varPackageVersionGroups.ContainsKey(VARPackageManifest.GetVPVGEKey(packagedRVGE.creatorName, packagedRVGE.packageName))) lastPackagedVersion = true;
            }

            if (rvge.localFileResource || lastPackagedVersion) parentBrowser.resourceSelectionsUI.DeSelectRVGE(rvge);
           
        }
        private void DeleteAllSelectedRVGEs()
        {
            List<ResourceVersionGroupEntry> rvgeList = new List<ResourceVersionGroupEntry>(parentBrowser.resourceSelectionsUI.GetResourceTypeSelections(resourceType, atomPresetResourceType).selectedRVGEs);
            foreach (var rvge in rvgeList)
            {
                DeleteRVGE(rvge);
            }
            parentBrowser.RefreshFilterResults();
        }

        private void DeletePrimaryRVGE()
        {
            DeleteRVGE(primaryRVGE);
            parentBrowser.RefreshFilterResults();
        }

        public void RefreshTargetAtomValue()
        {
            List<string> targetAtomNames = GetTargetAtomNames();
            ResetTargetAtomChoices(targetAtomNames);
 
            if (targetAtomJSMultiEnum.valType != JSONStorableMultiEnumStringChooser.topEnumValFlag)
            {
                if (targetAtomNames.Count == 0)
                {
                    if (resourceType == ResourceType.plugins) targetAtomJSMultiEnum.valTopEnum = PluginTargetTypes.sessionPlugins;
                    else targetAtomJSMultiEnum.valTopEnum = CreateNewAtom.newAtom;
                }
                else
                {                    
                    if (!targetAtomNames.Contains(targetAtomJSMultiEnum.mainVal))
                    {
                        SetFirstPreferredAtom(targetAtomNames);
                    }
                }
            }
            
        }

        private void SetFirstPreferredAtom(List<string> targetAtomNames)
        {
            var maleAtomNames = AtomUtils.GetPersonAtomNames(false, true);
            var femaleAtomNames = AtomUtils.GetPersonAtomNames(true, false);

            if (Settings.defaultTargetPersonTypeJSE.val == PersonGenderTypes.firstFemale && femaleAtomNames.Count > 0 && resoureTargetCategory == ResourceTargetCategory.personAtomTargetable)
            {
                if (targetAtomJSMultiEnum.mainVal != femaleAtomNames[0]) targetAtomJSMultiEnum.mainVal = femaleAtomNames[0];
            }
            else if (Settings.defaultTargetPersonTypeJSE.val == PersonGenderTypes.firstMale && maleAtomNames.Count > 0 && resoureTargetCategory == ResourceTargetCategory.personAtomTargetable)
            {
                if (targetAtomJSMultiEnum.mainVal != maleAtomNames[0]) targetAtomJSMultiEnum.mainVal = maleAtomNames[0];
            }
            else
            {
                if (targetAtomJSMultiEnum.mainVal != targetAtomNames[0]) targetAtomJSMultiEnum.mainVal = targetAtomNames[0];
            }
        }
        
        private List<string> GetTargetAtomNames()
        {
            List<string> targetAtomNames = new List<string>();
            if (resoureTargetCategory != ResourceTargetCategory.nonAtomTargetable)
            {
                List<Atom> atoms;

                if (resoureTargetCategory == ResourceTargetCategory.personAtomTargetable) atoms = AtomUtils.GetAtoms(incudeNonPersonAtoms: false);
                else if (resoureTargetCategory == ResourceTargetCategory.otherAtomTargetable) atoms = AtomUtils.GetAtoms(atomTypeFilter: atomPresetResourceType);
                else if (resoureTargetCategory == ResourceTargetCategory.subSceneTargetable) atoms = AtomUtils.GetAtoms(atomTypeFilter: "SubScene");
                else if (resoureTargetCategory == ResourceTargetCategory.femaleAtomTargetable) atoms = AtomUtils.GetPersonAtoms(true, false);
                else if (resoureTargetCategory == ResourceTargetCategory.maleAtomTargetable) atoms = AtomUtils.GetPersonAtoms(false, true);
                else if (resoureTargetCategory == ResourceTargetCategory.cuaTargetable) atoms = AtomUtils.GetAtoms(atomTypeFilter: "CustomUnityAsset");
                else atoms = AtomUtils.GetAtoms();

                foreach (Atom atom in atoms) targetAtomNames.Add(atom.name);
            }
            targetAtomNames.Sort();
            return targetAtomNames;
        }

        public void RefreshTargetAtomPopupCallback()
        {
            RefreshTargetAtomPopup();
        }
        public void RefreshTargetAtomPopup(bool forceAtomChoice = false)
        {
            var targetAtomNames = GetTargetAtomNames();
            ResetTargetAtomChoices(targetAtomNames);

            if (targetAtomNames.Count > 0 && forceAtomChoice && !targetAtomNames.Contains(targetAtomJSMultiEnum.mainVal))
            {
                SetFirstPreferredAtom(targetAtomNames);
            }
        }

        private void ResetTargetAtomChoices(List<string> atomNames)
        {
            if (resourceType == ResourceType.plugins) targetAtomJSMultiEnum.SetTopEnum(PluginTargetTypes.enumManifestName);
            else targetAtomJSMultiEnum.SetTopEnum(CreateNewAtom.enumManifestName);

            if (atomNames.Count == 0) targetAtomJSMultiEnum.SetMainChoices(null);         
            else targetAtomJSMultiEnum.SetMainChoices(atomNames);

            if (targetAtomJSMultiEnum.popup != null)
            {
                if (atomNames.Count > 10) targetAtomJSMultiEnum.popup.popupPanelHeight = 250f;
                else targetAtomJSMultiEnum.popup.popupPanelHeight = (atomNames.Count + 1) * 25f;
            }
            
        }
    }

    public class ResourceTagsSideUI: ResourceActionUIBase
    {

        private string tagNameFilter = "";

        private JSONStorableString tagFilterJSS;
        private JSONStorableEnumStringChooser tagCategoryFilterJSEnum;

        private ResourceTagsActionUI resourceTagsActionUI;

        GameObject tagListContentGO;
        private UIDynamicButton clearTagFilterButton;

        private ScrollRect scrollRect;
        private float scrollBarPosition = 1f;

        private Dictionary<JSONStorableBool, TriStateToggle> tagJSBtoTriStateToggle = new Dictionary<JSONStorableBool, TriStateToggle>();
        private Dictionary<JSONStorableBool, Tag> tagJSBtoTag = new Dictionary<JSONStorableBool, Tag>();

        private ResourceSelections resourceSelectionsUI { get { return parentBrowser.resourceSelectionsUI; } }
        private BrowserFiltersUI browserFiltersUI { get { return parentBrowser.browserFiltersUI; } }
        private int filtersResourceType { get { return browserFiltersUI.currentResourceType; } }
        private string filtersAtomPresetResourceType { get { return browserFiltersUI.currentAtomPresetResourceType; } }

        private int filtersSelectionMode { get { return browserFiltersUI.selectModeJSEnum.val; } }

        private bool secondTagSelectionBoxEnabled
        {
            get
            {
                return Settings.sideTagBoxEnabledJSB.val;
            }
            set
            {
                Settings.sideTagBoxEnabledJSB.val = value;
            }
        }

        public ResourceTagsSideUI(BrowserUI parentBrowser, ResourceTagsActionUI resourceTagsActionUI) : base (parentBrowser, parentBrowser.sideTagBoxGO)
        {
            this.resourceTagsActionUI= resourceTagsActionUI;

            float vSize = parentBrowser.screenHeight;
            if (Globals.isVREnabled) vSize = parentBrowser.screenHeight;

            RectTransform rt = parentBoxGO.GetComponent<RectTransform>();

            var scrollableContentGO = UI.CreateVScrollableContent(parentBoxGO, new Color(0.8f, 0.8f, 0.8f, 0.8f), AnchorPresets.vStretchLeft, 148f, rt.rect.height, new Vector2(2f, 0f));
            UI.ResetTopOffset(scrollableContentGO, -20f);
            scrollRect = scrollableContentGO.GetComponent<ScrollRect>();
            scrollRect.verticalScrollbar.onValueChanged.AddListener(delegate
            {
                if (scrollRect.verticalScrollbar.size > 0.999f) scrollBarPosition = 1f;
                else scrollBarPosition = scrollRect.verticalNormalizedPosition;
            });


            tagListContentGO = scrollableContentGO.transform.Find("scrollableContentViewportGO/vScrollableContentGO").gameObject;

            tagFilterJSS = new JSONStorableString("tagFilter", "");
            tagCategoryFilterJSEnum = new JSONStorableEnumStringChooser("tagCategoryFilter", TagCategory.enumManifestName, TagCategory.all, "Tag Category", TagCategoryFilterCallback);

            var textFieldUID = CreateTextField(parentBoxGO, tagFilterJSS, 70f, 20f, 18, 2, 0, AnchorPresets.topLeft, true, InputField.LineType.SingleLine, " Filter tags...", 0, 0, TagFilterNameCallback);
            parentBrowser.ApplyHelp(textFieldUID, "Enter text to filter the list of Tags that are available for this resource type.");

            clearTagFilterButton = CreateUIButton(parentBoxGO, 16f, 16f, "X", 18, 54f, -2f, AnchorPresets.topLeft);
            clearTagFilterButton.button.onClick.AddListener(delegate { tagFilterJSS.val = ""; RefreshUI(); });
            parentBrowser.ApplyHelp(clearTagFilterButton, "Clear the Tag search field of user entered text.");
            if (tagFilterJSS.val == "") clearTagFilterButton.gameObject.SetActive(false);

            var tagCategoryFilterUID = CreatePopup(parentBoxGO, tagCategoryFilterJSEnum, 60f, 20f, "", 20, 72f, 0f, AnchorPresets.topLeft, false, 0f);
            tagCategoryFilterUID.popup.onOpenPopupHandlers += RefreshTagCategoryFilter;

            var closeButton = CreateUIButton(parentBoxGO, 20f, 20f, "X", 20, 2f, 0f, AnchorPresets.topRight);
            closeButton.button.onClick.AddListener(delegate { resourceTagsActionUI.secondTagSelectionBoxEnabled = false; resourceTagsActionUI.RefreshUI(); });
        }

        public void RefreshTagCategoryFilter()
        {
            List<int> tagCategoryExclusions = new List<int>();
            if (resourceCategory != ResourceCategory.clothing)
            {
                tagCategoryExclusions.Add(TagCategory.bodyRegion);
                tagCategoryExclusions.Add(TagCategory.clothingType);
            }
            if (resourceCategory != ResourceCategory.preset)
            {
                tagCategoryExclusions.Add(TagCategory.gender);
            }
            if (resourceCategory != ResourceCategory.plugins)
            {
                tagCategoryExclusions.Add(TagCategory.pluginType);
            }
            tagCategoryFilterJSEnum.SetEnumChoices(TagCategory.enumManifestName, tagCategoryExclusions);
        }

        private void TagFilterNameCallback(string i)
        {
            tagNameFilter = TagManifest.NormalizeTagName(i);
            tagFilterJSS.val = tagNameFilter;
            tagFilterJSS.inputField.text = tagNameFilter;
            if (tagFilterJSS.val == "") clearTagFilterButton.gameObject.SetActive(false);
            else clearTagFilterButton.gameObject.SetActive(true);

            RePopulateTagSelectionsBox();
        }

        public void TagCategoryFilterCallback(int tagCategory)
        {
            RePopulateTagSelectionsBox();
        }

        public override void ClearUI()
        {
            parentBoxGO.SetActive(false);
        }

        public override void RefreshUI()
        {
            base.RefreshUI();
            if (secondTagSelectionBoxEnabled) parentBoxGO.SetActive(true);
            else parentBoxGO.SetActive(false);
        }

        public void RePopulateTagSelectionsBox()
        {
            float newScrollBarPosition = scrollBarPosition;

            foreach (var toggleKVP in tagJSBtoTriStateToggle)
            {

                GameObject.Destroy(toggleKVP.Value.toggleUID.gameObject);
                uiElements.Remove(toggleKVP.Value.toggleUID);
            }

            tagJSBtoTriStateToggle.Clear();
            tagJSBtoTag.Clear();

            List<Tag> availableTags = TagManifest.GetTags(filtersResourceType, filtersAtomPresetResourceType).Where(t => t.tagName.StartsWith(tagNameFilter) && (tagCategoryFilterJSEnum.val == TagCategory.all || tagCategoryFilterJSEnum.val == t.tagCategory)).ToList();

            if (availableTags.Count == 0) return;

            bool tagModificationDisabled = resourceCategory == ResourceCategory.clothing && !Globals.customFolderAccess;
            foreach (var tag in availableTags)
            {
                var tagJSB = new JSONStorableBool(tag.tagName, false, TagCallback);
                var tagToggle = CreateTriStateToggle(tagListContentGO, tagJSB, 148f-10f, 20f, tag.tagName, 20, 0, 0, AnchorPresets.middleLeft);
                if (tagModificationDisabled) tagToggle.toggle.interactable = false;

                RefreshTagToggle(tagToggle, tag);

                tagJSBtoTag.Add(tagJSB, tag);
                tagJSBtoTriStateToggle.Add(tagJSB, tagToggle);

                tagToggle.backgroundColor = TagCategory.GetColor(tag.tagCategory);

                TagPointerEnterExitHelp contextHelp = tagToggle.toggle.gameObject.AddComponent<TagPointerEnterExitHelp>();
                contextHelp.Init(tag, parentBrowser.helpTextJSS, resourceSelectionsUI.GetResourceTypeSelections(filtersResourceType, filtersAtomPresetResourceType), browserFiltersUI);
            }

            Globals.mvrScript.StartCoroutine(SetScollBarPosition(newScrollBarPosition));
        }

        private IEnumerator SetScollBarPosition(float newScrollBarPosition)
        {
            yield return new WaitForEndOfFrame();
            scrollRect.verticalNormalizedPosition = newScrollBarPosition;
        }

        public void TagCallback(JSONStorableBool valueJSB)
        {
            int state = resourceSelectionsUI.IsSelectionTagged(filtersResourceType, filtersAtomPresetResourceType, filtersSelectionMode, tagJSBtoTag[valueJSB]);

            int targetTriState = TriState.active;
            if (state == TriState.active) targetTriState = TriState.inactive;

            resourceSelectionsUI.SetSelectionTag(filtersResourceType, filtersAtomPresetResourceType, filtersSelectionMode, tagJSBtoTag[valueJSB], targetTriState);
//            tagJSBtoTriStateToggle[valueJSB].RefreshToggle(targetTriState);
            //            RefreshUI();
            resourceTagsActionUI.RefreshTagToggles();


            if (parentBrowser.browserFiltersUI.tagsFilterJSMultiEnum.valType != JSONStorableMultiEnumStringChooser.topEnumValFlag || parentBrowser.browserFiltersUI.tagsFilterJSMultiEnum.valTopEnum != TagGroups.anyTags) parentBrowser.RefreshFilterResults();
        }

        private void RefreshTagToggle(TriStateToggle tagToggle, Tag tag)
        {
            tagToggle.RefreshToggle(resourceSelectionsUI.IsSelectionTagged(filtersResourceType, filtersAtomPresetResourceType, filtersSelectionMode, tag));
            if (tag.tagCategory == TagCategory.combi)
            {
                tagToggle.toggle.interactable = false;
                tagToggle.toggleUID.textColor = Color.white;
            }
        }
        public void RefreshTagToggles()
        {
            foreach (var tagKVP in tagJSBtoTriStateToggle)
            {
                RefreshTagToggle(tagKVP.Value, tagJSBtoTag[tagKVP.Key]);
            }

        }
    }

    public class ResourceTagsActionUI : ResourceActionUIBase
    {
        private JSONStorableBool hiddenJSB;
        private JSONStorableBool favouriteJSB;
        private JSONStorableBool newJSB;

        private TriStateToggle hideToggle;
        private TriStateToggle favToggle;
        private TriStateToggle newToggle;

        private string tagNameFilter = "";

        private JSONStorableString tagFilterJSS;
        private JSONStorableEnumStringChooser tagCategoryFilterJSEnum;

        private ResourceSelections resourceSelectionsUI { get { return parentBrowser.resourceSelectionsUI; } }

        private BrowserFiltersUI browserFiltersUI { get { return parentBrowser.browserFiltersUI; } }

        private int filtersResourceType { get { return browserFiltersUI.currentResourceType; } }
        private string filtersAtomPresetResourceType { get { return browserFiltersUI.currentAtomPresetResourceType; } }

        private int filtersSelectionMode { get { return browserFiltersUI.selectModeJSEnum.val; } }

        private GameObject tagSelectionsScrollableBoxGO;
        private GameObject tagSelectionScrollableViewportContentGO;
        private List<GameObject> tagSelectionRows = new List<GameObject>();
        private UIDynamicButton createNewTagButton;
        private UIDynamicButton clearTagFilterButton;

        private ScrollRect scrollRect;
        private float scrollBarPosition =1f;

        private ResourceTagsSideUI resourceTagsSideUI;

        private bool tagExpansionModeEnabled =false;

        public bool secondTagSelectionBoxEnabled { get
            {
                return Settings.sideTagBoxEnabledJSB.val;
            }
            set
            {
                Settings.sideTagBoxEnabledJSB.val = value;
            }
        }

        private Dictionary<JSONStorableBool, TriStateToggle> tagJSBtoTriStateToggle = new Dictionary<JSONStorableBool, TriStateToggle>();
        private Dictionary<JSONStorableBool, Tag> tagJSBtoTag = new Dictionary<JSONStorableBool, Tag>();

        public ResourceTagsActionUI(BrowserUI _parentBrowser, GameObject _parentBoxGO) : base(_parentBrowser, _parentBoxGO)
        {
            hiddenJSB = new JSONStorableBool("hidden", false, HiddenCallback);
            favouriteJSB = new JSONStorableBool("favourite", false, FavouriteCallback);
            newJSB = new JSONStorableBool("new", false, NewCallback);
            tagFilterJSS = new JSONStorableString("tagFilter", "");
            tagCategoryFilterJSEnum = new JSONStorableEnumStringChooser("tagCategoryFilter", TagCategory.enumManifestName, TagCategory.all, "Tag Category",TagCategoryFilterCallback);

            resourceTagsSideUI = new ResourceTagsSideUI(parentBrowser, this);
        }

        public void TagCategoryFilterCallback(int tagCategory)
        {
            RePopulateTagSelectionsBox();
        }

        public void RefreshTagCategoryFilter()
        {
            List<int> tagCategoryExclusions = new List<int>();
            if (resourceCategory != ResourceCategory.clothing) {
                tagCategoryExclusions.Add(TagCategory.bodyRegion);
                tagCategoryExclusions.Add(TagCategory.clothingType);
            }
            if (resourceCategory != ResourceCategory.preset)
            {
                tagCategoryExclusions.Add(TagCategory.gender);
            }
            if (resourceCategory != ResourceCategory.plugins)
            {
                tagCategoryExclusions.Add(TagCategory.pluginType);
            }
            tagCategoryFilterJSEnum.SetEnumChoices(TagCategory.enumManifestName, tagCategoryExclusions);
        }

        public override void ClearUI()
        {
            base.ClearUI();
            GameObject.Destroy(tagSelectionsScrollableBoxGO);
            resourceTagsSideUI.ClearUI();
        }

        public void CancelTagExpansionMode()
        {
            tagExpansionModeEnabled = false;
        }

        public override void RefreshUI()
        {
            base.RefreshUI();

            ClearTagSelectionRows();

            tagJSBtoTriStateToggle.Clear();
            tagJSBtoTag.Clear();

            resourceTagsSideUI.RefreshUI();

            if (resourceSelectionsUI.SelectionCount(filtersResourceType, filtersAtomPresetResourceType) >0)
            {
                

                hideToggle = CreateTriStateToggle(parentBoxGO, hiddenJSB, 80, 20, "Hide", 20, -2, -3, anchorPreset: AnchorPresets.topRight);
                HiddenPointerEnterExitHelp hiddenContextHelp = hideToggle.toggle.gameObject.AddComponent<HiddenPointerEnterExitHelp>();
                hiddenContextHelp.Init(parentBrowser.helpTextJSS, resourceSelectionsUI.GetResourceTypeSelections(filtersResourceType, filtersAtomPresetResourceType), browserFiltersUI);

                favToggle = CreateTriStateToggle(parentBoxGO, favouriteJSB, 80, 20, "Favorite", 20, -2, -27, anchorPreset: AnchorPresets.topRight);
                FavPointerEnterExitHelp favContextHelp = favToggle.toggle.gameObject.AddComponent<FavPointerEnterExitHelp>();
                favContextHelp.Init(parentBrowser.helpTextJSS, resourceSelectionsUI.GetResourceTypeSelections(filtersResourceType, filtersAtomPresetResourceType), browserFiltersUI);

                newToggle = CreateTriStateToggle(parentBoxGO, newJSB, 80, 20, "New\\Updated", 20, -2, -50, anchorPreset: AnchorPresets.topRight);
                NewPointerEnterExitHelp newContextHelp = newToggle.toggle.gameObject.AddComponent<NewPointerEnterExitHelp>();
                newContextHelp.Init(parentBrowser.helpTextJSS, resourceSelectionsUI.GetResourceTypeSelections(filtersResourceType, filtersAtomPresetResourceType), browserFiltersUI);               

                if (!Globals.addonPackageFilePrefsFolderAccess || !Globals.customFolderAccess || !Globals.scenesFolderAccess)
                {
                    favToggle.toggle.interactable = false;
                    hideToggle.toggle.interactable = false;

                    parentBrowser.ApplyHelp(hideToggle.toggleUID, "Enable BA Sym Links using .bat file provided in PluginData folder to allow toggling the VAM Hidden state from within Browser Assist. See BA VAMHub page for detailed instructions.");
                    parentBrowser.ApplyHelp(favToggle.toggleUID, "Enable BA Sym Links using .bat file provided in PluginData folder to allow toggling the VAM Favorite state from within Browser Assist. See BA VAMHub page for detailed instructions.");
                }

                hideToggle.RefreshToggle(resourceSelectionsUI.IsSelectionHidden(filtersResourceType, filtersAtomPresetResourceType, filtersSelectionMode));

                favToggle.RefreshToggle(resourceSelectionsUI.IsSelectionFavourite(filtersResourceType, filtersAtomPresetResourceType, filtersSelectionMode));

                newToggle.RefreshToggle(resourceSelectionsUI.IsSelectionNew(filtersResourceType, filtersAtomPresetResourceType, filtersSelectionMode));

                var textFieldUID = CreateTextField(parentBoxGO, tagFilterJSS, 150f, 20f, 20, 2, -2, AnchorPresets.topLeft, true, InputField.LineType.SingleLine, " Filter or create new tag...", 0, 0,TagFilterNameCallback);
                parentBrowser.ApplyHelp(textFieldUID, "Enter text to filter the list of Tags that are available for this resource type. Create new Tags, where an existing Tag does not match the entered search text.");

                clearTagFilterButton = CreateUIButton(parentBoxGO, 16f, 16f, "X", 20, 134f, -4, AnchorPresets.topLeft);
                clearTagFilterButton.button.onClick.AddListener(delegate { tagFilterJSS.val=""; RefreshUI(); });
                parentBrowser.ApplyHelp(clearTagFilterButton, "Clear the Tag search field of user entered text.");
                if (tagFilterJSS.val == "") clearTagFilterButton.gameObject.SetActive(false);

                createNewTagButton = CreateUIButton(parentBoxGO, 80f, 20f, "Create Tag", 22, 155, -2, AnchorPresets.topLeft);
                createNewTagButton.buttonColor = Color.green;
                createNewTagButton.textColor = Color.black;
                RefreshNewTagButton();
                bool tagModificationDisabled = resourceCategory == ResourceCategory.clothing && !Globals.customFolderAccess;
                if (tagModificationDisabled) createNewTagButton.button.interactable = false;

                createNewTagButton.button.onClick.AddListener(delegate {
                    var newTag = TagManifest.CreateTag(tagNameFilter, TagCategory.userDefined, filtersResourceType, filtersAtomPresetResourceType);
                    resourceSelectionsUI.SetSelectionTag(filtersResourceType, filtersAtomPresetResourceType, filtersSelectionMode, newTag, TriState.active);
                    RefreshUI();
                });
                parentBrowser.ApplyHelp(createNewTagButton, "Create a new Tag and assign it to the selected resources.");

                RefreshTagCategoryFilter();
                var tagCategoryFilterUID = CreatePopup(parentBoxGO, tagCategoryFilterJSEnum, 120f, 20f, "Tag Filter", 20, tagSelectionBoxWidth - 120f-24f-20f, -2f,AnchorPresets.topLeft,true,0.4f);
                tagCategoryFilterUID.popup.onOpenPopupHandlers += RefreshTagCategoryFilter;


                float verticalOffset = -24f;
                float secondaryTagBoxEnabledButtonOffset = -2f;


                if (tagExpansionModeEnabled)
                {
                    verticalOffset = -2f;
                    secondaryTagBoxEnabledButtonOffset = 18f;
                    var tagExpandButton = CreateUIButton(parentBoxGO, 20f, 20f, "v", 20, tagSelectionBoxWidth-18f, 18f, AnchorPresets.topLeft);
                    tagExpandButton.button.onClick.AddListener(delegate {
                        tagExpansionModeEnabled = false;
                        RefreshUI();
                    });
                    parentBrowser.ApplyHelp(tagExpandButton, "Shrink the size of the tagging window (restores the tag search filter and the favorite, hide and new/updated tag check boxes).");
                    
                }
                else
                {
                    var tagExpandButton = CreateUIButton(parentBoxGO, 20f, 20f, "^", 26, tagSelectionBoxWidth - 20f, -2f, AnchorPresets.topLeft);
                    tagExpandButton.button.onClick.AddListener(delegate {
                        tagExpansionModeEnabled = true;
                        RefreshUI();
                    });
                    parentBrowser.ApplyHelp(tagExpandButton, "Expand the size of the tagging window (disables the tag search filter and the favorite, hide and new/updated tag check boxes).");

                    
                    
                }

                UIDynamicButton secondaryTagBox;
                if (secondTagSelectionBoxEnabled)
                {
                    secondaryTagBox = CreateUIButton(parentBoxGO, 20f, 20f, ">", 26, tagSelectionBoxWidth - 40f, secondaryTagBoxEnabledButtonOffset, AnchorPresets.topLeft);
                    parentBrowser.ApplyHelp(secondaryTagBox, "Expand the size of the tagging window (disables the tag search filter and the favorite, hide and new/updated tag check boxes).");
                }
                else
                {
                    secondaryTagBox = CreateUIButton(parentBoxGO, 20f, 20f, "<", 26, tagSelectionBoxWidth - 40f, secondaryTagBoxEnabledButtonOffset, AnchorPresets.topLeft);
                    parentBrowser.ApplyHelp(secondaryTagBox, "Expand the size of the tagging window (disables the tag search filter and the favorite, hide and new/updated tag check boxes).");
                }

                secondaryTagBox.button.onClick.AddListener(delegate {
                    secondTagSelectionBoxEnabled = !secondTagSelectionBoxEnabled;
                    RefreshUI();
                });


                tagSelectionsScrollableBoxGO = UI.CreateVScrollableContent(parentBoxGO, new Color(0.5f, 0.5f, 0.65f, 0.98f), AnchorPresets.topLeft, tagSelectionBoxWidth, tagSelectionBoxHeight, new Vector2(2f, verticalOffset));
                tagSelectionScrollableViewportContentGO = tagSelectionsScrollableBoxGO.transform.Find("scrollableContentViewportGO/vScrollableContentGO").gameObject;

                RePopulateTagSelectionsBox();

                scrollRect = tagSelectionsScrollableBoxGO.GetComponent<ScrollRect>();
                scrollRect.verticalScrollbar.onValueChanged.AddListener(delegate
                {
                    if (scrollRect.verticalScrollbar.size > 0.999f) scrollBarPosition = 1f;
                    else scrollBarPosition = scrollRect.verticalNormalizedPosition;
                });

                SetScollBarPosition(scrollBarPosition);            
            }

        }

        private void RefreshNewTagButton()
        {
            if (tagNameFilter == "hide" || tagNameFilter == "favorite" || tagNameFilter.EndsWith(" ") || tagNameFilter =="" || TagManifest.GetTags(filtersResourceType,filtersAtomPresetResourceType).Select(x => x.tagName).Contains(tagNameFilter)) createNewTagButton.gameObject.SetActive(false);
            else createNewTagButton.gameObject.SetActive(true);
        }

        private void TagFilterNameCallback(string i)
        {
            tagNameFilter = TagManifest.NormalizeTagName(i);
            tagFilterJSS.val = tagNameFilter;
            tagFilterJSS.inputField.text = tagNameFilter;
            if (tagFilterJSS.val == "") clearTagFilterButton.gameObject.SetActive(false);
            else clearTagFilterButton.gameObject.SetActive(true);

            RefreshNewTagButton();
            RePopulateTagSelectionsBox();

        }
        private void ClearTagSelectionRows()
        {
            foreach (var rowGO in tagSelectionRows) GameObject.Destroy(rowGO);
            tagSelectionRows.Clear();
        }

        private float tagSelectionBoxWidth { get
            {
                if (tagExpansionModeEnabled) return 638f;
                else return 550f;
            } }

        private float tagSelectionBoxHeight
        {
            get
            {
                if (tagExpansionModeEnabled) return 70f;
                else return 48f;
            }
        }
        private GameObject CreateSelectTagRow()
        {
            GameObject currentRowGO = UI.AddChildGOImage(tagSelectionScrollableViewportContentGO, new Color(0f, 0.8f, 0f, 0f), AnchorPresets.hStretchTop, tagSelectionBoxWidth, 20f, Vector2.zero);
            tagSelectionRows.Add(currentRowGO);

            return currentRowGO;
        }

        private float GetTagButtonWidth(Tag tag)
        {
            float length = tag.tagName.Length;

            foreach (char c in tag.tagName)
            {
                if (c == 'm' || c == 'w') length += 0.5f;
                if (c == 'i' || c == 'j' || c == 't' || c == 'f' || c == 'l') length -= 0.5f;
            }

            return Mathf.Clamp((length * 5)+25, 40f, 150f);

        }

        private void RefreshTagToggle(TriStateToggle tagToggle, Tag tag)
        {
            tagToggle.RefreshToggle(resourceSelectionsUI.IsSelectionTagged(filtersResourceType, filtersAtomPresetResourceType, filtersSelectionMode, tag));
            if (tag.tagCategory == TagCategory.combi)
            {
                tagToggle.toggle.interactable = false;
                tagToggle.toggleUID.textColor = Color.white;
            }
        }

        public void RefreshTagToggles()
        {
            foreach (var tagKVP in tagJSBtoTriStateToggle)
            {
                RefreshTagToggle(tagKVP.Value, tagJSBtoTag[tagKVP.Key]);
            }
            hideToggle.RefreshToggle(resourceSelectionsUI.IsSelectionHidden(filtersResourceType, filtersAtomPresetResourceType, filtersSelectionMode));

            favToggle.RefreshToggle(resourceSelectionsUI.IsSelectionFavourite(filtersResourceType, filtersAtomPresetResourceType, filtersSelectionMode));

            newToggle.RefreshToggle(resourceSelectionsUI.IsSelectionNew(filtersResourceType, filtersAtomPresetResourceType, filtersSelectionMode));

            if (resourceTagsSideUI != null) resourceTagsSideUI.RefreshTagToggles();
        }

        public void RePopulateTagSelectionsBox()
        {
            float newScrollBarPosition = scrollBarPosition;
            
            foreach (var toggleKVP in tagJSBtoTriStateToggle)
            {

                GameObject.Destroy(toggleKVP.Value.toggleUID.gameObject);
                uiElements.Remove(toggleKVP.Value.toggleUID);
            }

            ClearTagSelectionRows();

            tagJSBtoTriStateToggle.Clear();
            tagJSBtoTag.Clear();

            float currentRowTagsWidth = 0f;

            List<Tag> availableTags = TagManifest.GetTags(filtersResourceType, filtersAtomPresetResourceType).Where(t => t.tagName.StartsWith(tagNameFilter) && (tagCategoryFilterJSEnum.val == TagCategory.all || tagCategoryFilterJSEnum.val == t.tagCategory)).ToList();

            if (resourceTagsSideUI != null) resourceTagsSideUI.RePopulateTagSelectionsBox();

            if (availableTags.Count == 0) return;

            bool tagModificationDisabled = resourceCategory == ResourceCategory.clothing && !Globals.customFolderAccess;

            GameObject currentRowGO = CreateSelectTagRow();
            foreach (var tag in availableTags)
            {
                float nextButtonWidth = GetTagButtonWidth(tag);

                float maxWidthBorder = 0f;
                if (Globals.isVREnabled) maxWidthBorder = 25f;

                if (currentRowTagsWidth + nextButtonWidth  > (tagSelectionBoxWidth - maxWidthBorder))
                {
                    currentRowGO = CreateSelectTagRow();
                    currentRowTagsWidth = 0f;
                }
                var tagJSB = new JSONStorableBool(tag.tagName, false, TagCallback);
                var tagToggle = CreateTriStateToggle(currentRowGO, tagJSB, nextButtonWidth, 20f, tag.tagName, 20, currentRowTagsWidth, 0, AnchorPresets.middleLeft);

                RefreshTagToggle(tagToggle, tag);
                if (tagModificationDisabled) tagToggle.toggle.interactable = false;

                tagJSBtoTag.Add(tagJSB, tag);
                tagJSBtoTriStateToggle.Add(tagJSB, tagToggle);

                tagToggle.backgroundColor = TagCategory.GetColor(tag.tagCategory);

               TagPointerEnterExitHelp contextHelp = tagToggle.toggle.gameObject.AddComponent<TagPointerEnterExitHelp>();
                contextHelp.Init(tag, parentBrowser.helpTextJSS, resourceSelectionsUI.GetResourceTypeSelections(filtersResourceType, filtersAtomPresetResourceType), browserFiltersUI);

                currentRowTagsWidth += nextButtonWidth;

                currentRowTagsWidth += 3f;
            }

            
            Globals.mvrScript.StartCoroutine(SetScollBarPosition(newScrollBarPosition));
        }
        private IEnumerator SetScollBarPosition(float newScrollBarPosition)
        {
            yield return new WaitForEndOfFrame();
            scrollRect.verticalNormalizedPosition = newScrollBarPosition;
        }

        public void TagCallback(JSONStorableBool valueJSB)
        {
            int state = resourceSelectionsUI.IsSelectionTagged(filtersResourceType, filtersAtomPresetResourceType, filtersSelectionMode, tagJSBtoTag[valueJSB]);

            int targetTriState = TriState.active;
            if (state == TriState.active) targetTriState = TriState.inactive;

            resourceSelectionsUI.SetSelectionTag(filtersResourceType, filtersAtomPresetResourceType, filtersSelectionMode, tagJSBtoTag[valueJSB], targetTriState);
//            tagJSBtoTriStateToggle[valueJSB].RefreshToggle(targetTriState);
 //           RefreshUI();
            RefreshTagToggles();

            if (parentBrowser.browserFiltersUI.tagsFilterJSMultiEnum.valType!=JSONStorableMultiEnumStringChooser.topEnumValFlag || parentBrowser.browserFiltersUI.tagsFilterJSMultiEnum.valTopEnum!=TagGroups.anyTags) parentBrowser.RefreshFilterResults();
        }

        public void HiddenCallback(bool value)
        {
            int state = resourceSelectionsUI.IsSelectionHidden(filtersResourceType, filtersAtomPresetResourceType, filtersSelectionMode);

            if (state == TriState.active)
            {
                resourceSelectionsUI.SetSelectionHidden(filtersResourceType, filtersAtomPresetResourceType, filtersSelectionMode, TriState.inactive);

                hideToggle.RefreshToggle(resourceSelectionsUI.IsSelectionHidden(filtersResourceType, filtersAtomPresetResourceType, filtersSelectionMode));
            }
            else
            {
                resourceSelectionsUI.SetSelectionHidden(filtersResourceType, filtersAtomPresetResourceType, filtersSelectionMode, TriState.active);
                hideToggle.RefreshToggle(resourceSelectionsUI.IsSelectionHidden(filtersResourceType, filtersAtomPresetResourceType, filtersSelectionMode));
            }

            RefreshUI();
            if (!parentBrowser.browserFiltersUI.showHiddenJSB.val)parentBrowser.RefreshFilterResults();


        }

        public void FavouriteCallback(bool value)
        {
            int state = resourceSelectionsUI.IsSelectionFavourite(filtersResourceType, filtersAtomPresetResourceType, filtersSelectionMode);

            if (state == TriState.active)
            {
                resourceSelectionsUI.SetSelectionFavourite(filtersResourceType, filtersAtomPresetResourceType, filtersSelectionMode, TriState.inactive);
                favToggle.RefreshToggle(TriState.inactive);
            }
            else
            {
                resourceSelectionsUI.SetSelectionFavourite(filtersResourceType, filtersAtomPresetResourceType, filtersSelectionMode, TriState.active);
                favToggle.RefreshToggle(TriState.active);
            }

            RefreshUI();
            if (parentBrowser.browserFiltersUI.onlyFavouritesJSB.val) parentBrowser.RefreshFilterResults();
        }

        public void NewCallback(bool value)
        {
            int state = resourceSelectionsUI.IsSelectionNew(filtersResourceType,filtersAtomPresetResourceType, filtersSelectionMode);

            if (state == TriState.active)
            {
                resourceSelectionsUI.SetSelectionNew(filtersResourceType, filtersAtomPresetResourceType, filtersSelectionMode, TriState.inactive);
                newToggle.RefreshToggle(TriState.inactive);
            }
            else
            {
                resourceSelectionsUI.SetSelectionNew(filtersResourceType, filtersAtomPresetResourceType, filtersSelectionMode, TriState.active);
                newToggle.RefreshToggle(TriState.active);
            }
            RefreshUI();
            if (parentBrowser.browserFiltersUI.onlyNewJSB.val) parentBrowser.RefreshFilterResults();
        }

    }

    public class ResourcePositionActionUI : ResourceActionUIBase
    {

        private bool isNewAtomCreation { get { return resourceActionTab.targetAtomJSMultiEnum.valType == JSONStorableMultiEnumStringChooser.topEnumValFlag; } }


        private ResourcePrimaryActionUI resourceActionTab;

        private GameObject rotationLabelGO;
        private GameObject positionLabelGO;

        public static AtomPositionRotationSync existingAtomPosition ;
        public static RelativeAtomPositionRotation newAtomPosition ;
        public ResourcePositionActionUI(BrowserUI _parentBrowser, GameObject _parentBoxGO, ResourcePrimaryActionUI _resourceActionTab) : base(_parentBrowser, _parentBoxGO)
        {
            resourceActionTab = _resourceActionTab;

            newAtomPosition.Init(RelativeAbsoluteChangeCallback, RelativeToAtomCallback, RelativeToAtomCallback,true);

            
            if (Globals.isVREnabled) newAtomPosition.CopyFrom(Settings.vrDefaultNewAtomPosRot);
            else newAtomPosition.CopyFrom(Settings.desktopDefaultNewAtomPosRot);
        }

        public override void AtomNameUpdated(string oldName, string newName)
        {
            if (newAtomPosition!=null) newAtomPosition.AtomNameUpdated(oldName, newName);  
        }

        public override void OnAtomRemoved(Atom atom)
        {
            if (newAtomPosition != null) newAtomPosition.OnAtomRemoved(atom);
        }

        private void RelativeAbsoluteChangeCallback(int val)
        {
            RefreshUI();
        }

        private void RelativeToAtomCallback(int val)
        {
            RefreshUI();
        }

        private void RelativeToAtomCallback(string val)
        {
            RefreshUI();
        }

        public override void ClearUI()
        {
            base.ClearUI();
            
            newAtomPosition.SetGuideAtomVisible(false);
            if (rotationLabelGO != null) GameObject.Destroy(rotationLabelGO);
            if (positionLabelGO != null) GameObject.Destroy(positionLabelGO);

            if (posOptionsPopupGO!=null) GameObject.Destroy(posOptionsPopupGO);
            if (rotOptionsPopupGO!=null) GameObject.Destroy(rotOptionsPopupGO);
            currentPositionRotationAxis = -1;
        }

        public override void RefreshUI()
        {
            base.RefreshUI();
            // Target Atom
            if (resourceCategory == ResourceCategory.preset || resourceType == ResourceType.plugins || resourceType == ResourceType.subScenes || resourceCategory == ResourceCategory.clothing || resourceType == ResourceType.cua)
            {
                resourceActionTab.targetAtomJSMultiEnum.setTopEnumCallbackFunction = TargetAtomTopEnumCallBack;
                resourceActionTab.targetAtomJSMultiEnum.setMainValCallbackFunction = TargetAtomMainEnumCallBack;

                var targetAtomPopup = CreateScrollablePopup(parentBoxGO, resourceActionTab.targetAtomJSMultiEnum, 160f, 25f, "Target Atom", 20, edgeWidth, -edgeWidth, 500f, true, true, AnchorPresets.topLeft);
                targetAtomPopup.popup.onOpenPopupHandlers += resourceActionTab.RefreshTargetAtomPopupCallback;
                parentBrowser.ApplyHelp(targetAtomPopup, "Select a target atom from the current scene to reposition. Select '<New Atom>' to specify the position of the next new atom to be created.");
                resourceActionTab.RefreshTargetAtomValue();

                AtomPositionRotationSync currentAtomPosition = existingAtomPosition;

                float relAbsLabelProportion = 0.4f;
                if (isNewAtomCreation )
                {
                    newAtomPosition.SetGuideAtomVisible(true);
                    currentAtomPosition = newAtomPosition;
                    if (newAtomPosition.isRelativePosition || newAtomPosition.isRelativeRotation)
                    {
                        var relativeAtomPopup = CreateScrollablePopup(parentBoxGO, newAtomPosition.relativeAtomNameJSME, 160f, 25f, "Relative to Atom",18,edgeWidth,-edgeWidth-25f,500f,true,true,AnchorPresets.topLeft,20);
                        relativeAtomPopup.popup.onOpenPopupHandlers += RefreshRelativeAtomPopup;
                        parentBrowser.ApplyHelp(relativeAtomPopup, "New atoms will be created relative to this atom (or VR/Dekstop control point). Only the position and rotation elements marked as REL will be relative to this atom. Atom creation can also be relative to the VR Headset, Controllers or the main desktop view (ie. in front of you).");


                        if (newAtomPosition.relativeAtomNameJSME.valType == JSONStorableMultiEnumStringChooser.mainValFlag)
                        {
                            var relativeAtom = SuperController.singleton.GetAtomByUid(newAtomPosition.relativeAtomNameJSME.mainVal);
                            if (relativeAtom !=null && relativeAtom.type == "Person")
                            {
                                var relativeNodePopup = CreateScrollablePopup(parentBoxGO, newAtomPosition.relativeNodeNameJSSC, 160f, 25f, "Relative to Node", 18, edgeWidth, -edgeWidth - 50f, 500f, true, true, AnchorPresets.topLeft, 20);
                                relativeNodePopup.popup.onOpenPopupHandlers += RefreshRelativeNodePopup;
                                parentBrowser.ApplyHelp(relativeNodePopup, "New atoms will be created relative to this atom node. Only the position and rotation elements marked as REL will be relative to this contoller.");
                            }
                            
                        }
                        
                    }
                    
                }
                else
                {
                    currentAtomPosition.SetAtomName(resourceActionTab.targetAtomJSMultiEnum.mainVal);
                    relAbsLabelProportion = 1f;
                    if (currentAtomPosition.baAtomCollisionEnabledJSB != null)
                    {
                        var collisionToggle = CreateToggle(parentBoxGO, currentAtomPosition.baAtomCollisionEnabledJSB, 160f, 20f, "Collision Enabled", 20, edgeWidth, -edgeWidth - 25f, AnchorPresets.topLeft);
                        parentBrowser.ApplyHelp(collisionToggle, "Collision for the target atom can be toggled on or off to facilitate easier re-positioning.");
                    }
                    
                }

                var posXRelAbsPopup = CreatePopup(parentBoxGO, newAtomPosition.posXRelAbsJSE, 60, 20, "X", 24, edgeWidth + 165f, -edgeWidth - 10, AnchorPresets.topLeft, true, relAbsLabelProportion);
                parentBrowser.ApplyHelp(posXRelAbsPopup, "Choose whether the X axis position for a new Atom is specified relative (Rel) to another position or if it's an absolute (Abs) Global position in World Space.");

                var posYRelAbsPopup = CreatePopup(parentBoxGO, newAtomPosition.posYRelAbsJSE, 60, 20, "Y", 24, edgeWidth + 165f, -edgeWidth - 10f - 20f, AnchorPresets.topLeft, true, relAbsLabelProportion);
                parentBrowser.ApplyHelp(posYRelAbsPopup, "Choose whether the Y axis position (height) for a new Atom is specified relative (Rel) to another position or if it's an absolute (Abs) Global position in World Space.");

                var posZRelAbsPopup = CreatePopup(parentBoxGO, newAtomPosition.posZRelAbsJSE, 60, 20, "Z", 24, edgeWidth + 165f, -edgeWidth - 10f - 40f, AnchorPresets.topLeft, true, relAbsLabelProportion);
                parentBrowser.ApplyHelp(posZRelAbsPopup, "Choose whether the Z axis position for a new Atom is specified relative (Rel) to another position or if it's an absolute (Abs) Global position in World Space.");

                var rotXRelAbsPopup = CreatePopup(parentBoxGO, newAtomPosition.rotXRelAbsJSE, 60, 20, "X", 24, -160f - edgeWidth, -edgeWidth - 10, AnchorPresets.topRight, true, relAbsLabelProportion);
                parentBrowser.ApplyHelp(rotXRelAbsPopup, "Choose whether the X axis rotation (pitch) for a new Atom is specified relative (Rel) to another rotation or if it's an absolute (Abs) Global rotation in World Space.");

                var rotYRelAbsPopup = CreatePopup(parentBoxGO, newAtomPosition.rotYRelAbsJSE, 60, 20, "Y", 24, -160f - edgeWidth, -edgeWidth - 10f - 20f, AnchorPresets.topRight, true, relAbsLabelProportion);
                parentBrowser.ApplyHelp(rotYRelAbsPopup, "Choose whether the Y axis rotation (yaw) for a new Atom is specified relative (Rel) to another rotation or if it's an absolute (Abs) Global rotation in World Space.");

                var rotZRelAbsPopup = CreatePopup(parentBoxGO, newAtomPosition.rotZRelAbsJSE, 60, 20, "Z", 24, -160f - edgeWidth, -edgeWidth - 10f - 40f, AnchorPresets.topRight, true, relAbsLabelProportion);
                parentBrowser.ApplyHelp(rotZRelAbsPopup, "Choose whether the Z axis rotation (roll) for a new Atom is specified relative (Rel) to another rotation or if it's an absolute (Abs) Global rotation in World Space.");

                positionLabelGO = UI.CreateBasicTextBox(parentBoxGO, new JSONStorableString("", "POSITION"), 100f, 12f, 20, edgeWidth + 165f, -1f, AnchorPresets.topLeft, true, backgoundColor: BrowserUI.darkBlueGrey);
                rotationLabelGO= UI.CreateBasicTextBox(parentBoxGO, new JSONStorableString("", "ROTATION"), 100f, 12f, 20, -120f, -1f, AnchorPresets.topRight, true, backgoundColor: BrowserUI.darkBlueGrey);

                
                var targetAtomPositionXSlider = CreateSlider(parentBoxGO, currentAtomPosition.positionXJSF, 165f,20f,"Postion X",26,edgeWidth+165f+60f,-edgeWidth-10f,  AnchorPresets.topLeft, 0f, 0f, 45f, true, true);
                targetAtomPositionXSlider.valueFormat = "F3";
                parentBrowser.ApplyHelp(targetAtomPositionXSlider, "Adjust the X axis position for the target atom.");
                var xPosOptionsButton = CreateUIButton(parentBoxGO, 20f, 20f, "^", 26, edgeWidth + 165f + 60f + 165f, -edgeWidth - 10f, AnchorPresets.topLeft);
                xPosOptionsButton.button.onClick.AddListener(delegate
                {
                    ToggleAxisOptionsPopup(PositionRotationAxis.xPosition);
                });
                parentBrowser.ApplyHelp(xPosOptionsButton, "Toggle options menu for the X Position axis slider");

                var targetAtomPositionYSlider = CreateSlider(parentBoxGO, currentAtomPosition.positionYJSF, 165f, 20f, "Postion Y", 26, edgeWidth + 165f + 60f, -edgeWidth - 10f - 20f, AnchorPresets.topLeft, 0f, 0f, 45f, true, true);
                targetAtomPositionYSlider.valueFormat = "F3";
                parentBrowser.ApplyHelp(targetAtomPositionYSlider, "Adjust the Y axis position for the target atom.");
                var yPosOptionsButton = CreateUIButton(parentBoxGO, 20f, 20f, "^", 26, edgeWidth + 165f + 60f + 165f, -edgeWidth - 10f-20f, AnchorPresets.topLeft);
                yPosOptionsButton.button.onClick.AddListener(delegate
                {
                    ToggleAxisOptionsPopup(PositionRotationAxis.yPosition);
                });
                parentBrowser.ApplyHelp(yPosOptionsButton, "Toggle options menu for the Y Position axis slider");


                var targetAtomPositionZSlider = CreateSlider(parentBoxGO, currentAtomPosition.positionZJSF, 165f, 20f, "Postion Z", 26, edgeWidth + 165f + 60f, -edgeWidth - 10f - 40f, AnchorPresets.topLeft, 0f, 0f, 45f, true, true);
                targetAtomPositionZSlider.valueFormat = "F3";
                parentBrowser.ApplyHelp(targetAtomPositionZSlider, "Adjust the Z axis position for the target atom.");
                var zPosOptionsButton = CreateUIButton(parentBoxGO, 20f, 20f, "^", 26, edgeWidth + 165f + 60f + 165f, -edgeWidth - 10f - 40f, AnchorPresets.topLeft);
                zPosOptionsButton.button.onClick.AddListener(delegate
                {
                    ToggleAxisOptionsPopup(PositionRotationAxis.zPosition);
                });
                parentBrowser.ApplyHelp(zPosOptionsButton, "Toggle options menu for the Z Position axis slider");

                var targetAtomRotationXSlider = CreateSlider(parentBoxGO, currentAtomPosition.rotationXJSF, 140f, 20f, "Rotation X", 24, -edgeWidth-20f, -edgeWidth-10f, AnchorPresets.topRight, 0f, 0f, 35f,true);
                targetAtomRotationXSlider.valueFormat = "F1";
                parentBrowser.ApplyHelp(targetAtomRotationXSlider, "Adjust the X axis rotation (Pitch) for the target atom.");
                var xRotOptionsButton = CreateUIButton(parentBoxGO, 20f, 20f, "^", 26, -edgeWidth, -edgeWidth - 10f, AnchorPresets.topRight);
                xRotOptionsButton.button.onClick.AddListener(delegate
                {
                    ToggleAxisOptionsPopup(PositionRotationAxis.xRotation);
                });
                parentBrowser.ApplyHelp(xRotOptionsButton, "Toggle options menu for the X Rotation axis slider");

                var targetAtomRotationYSlider = CreateSlider(parentBoxGO, currentAtomPosition.rotationYJSF, 140f, 20f, "Rotation Y", 24, -edgeWidth-20f, -edgeWidth - 10f-20f, AnchorPresets.topRight, 0f, 0f, 35f, true);
                targetAtomRotationYSlider.valueFormat = "F1";
                parentBrowser.ApplyHelp(targetAtomRotationYSlider, "Adjust the Y axis rotation (Yaw) for the target atom.");
                var yRotOptionsButton = CreateUIButton(parentBoxGO, 20f, 20f, "^", 26,  -edgeWidth, -edgeWidth - 10f-20f, AnchorPresets.topRight);
                yRotOptionsButton.button.onClick.AddListener(delegate
                {
                    ToggleAxisOptionsPopup(PositionRotationAxis.yRotation);
                });
                parentBrowser.ApplyHelp(yRotOptionsButton, "Toggle options menu for the Y Rotation axis slider");

                var targetAtomRotationZSlider = CreateSlider(parentBoxGO, currentAtomPosition.rotationZJSF, 140f, 20f, "Rotation Z", 24, -edgeWidth-20f, -edgeWidth - 10f-40f, AnchorPresets.topRight, 0f, 0f, 35f, true);
                targetAtomRotationZSlider.valueFormat = "F1";
                parentBrowser.ApplyHelp(targetAtomRotationZSlider, "Adjust the Z axis rotation (Roll) for the target atom.");
                var zRotOptionsButton = CreateUIButton(parentBoxGO, 20f, 20f, "^", 26, -edgeWidth, -edgeWidth - 10f-40f, AnchorPresets.topRight);
                zRotOptionsButton.button.onClick.AddListener(delegate
                {
                    ToggleAxisOptionsPopup(PositionRotationAxis.zRotation);
                });
                parentBrowser.ApplyHelp(zRotOptionsButton, "Toggle options menu for the Z Rotation axis slider");


            }
            CreatePostionRotationOptionPopups();
        }

        private int currentPositionRotationAxis = -1;
        private GameObject posOptionsPopupGO = null;
        private GameObject rotOptionsPopupGO = null;

        private void ToggleAxisOptionsPopup(int positionRotationAxis)
        {
            posOptionsPopupGO.SetActive(false);
            rotOptionsPopupGO.SetActive(false);
            if (currentPositionRotationAxis== positionRotationAxis)
            {
                
                currentPositionRotationAxis = -1;
            }
            else
            {
                RectTransform rtPos = posOptionsPopupGO.GetComponent<RectTransform>();
                RectTransform rtRot = rotOptionsPopupGO.GetComponent<RectTransform>();
                if (positionRotationAxis == PositionRotationAxis.xPosition)
                {
                    posOptionsPopupGO.SetActive(true);
                    rtPos.anchoredPosition = new Vector2(390f, 60f);
                }
                else if (positionRotationAxis == PositionRotationAxis.yPosition)
                {
                    posOptionsPopupGO.SetActive(true);
                    rtPos.anchoredPosition = new Vector2(390f, 60f-20f);
                }
                else if (positionRotationAxis == PositionRotationAxis.zPosition)
                {
                    posOptionsPopupGO.SetActive(true);
                    rtPos.anchoredPosition = new Vector2(390f, 60f-40f);
                }
                else if (positionRotationAxis == PositionRotationAxis.xRotation)
                {
                    rotOptionsPopupGO.SetActive(true);
                    rtRot.anchoredPosition = new Vector2(570f, 60f);
                }
                else if (positionRotationAxis == PositionRotationAxis.yRotation)
                {
                    rotOptionsPopupGO.SetActive(true);
                    rtRot.anchoredPosition = new Vector2(570f, 60f - 20f);
                }
                else if (positionRotationAxis == PositionRotationAxis.zRotation)
                {
                    rotOptionsPopupGO.SetActive(true);
                    rtRot.anchoredPosition = new Vector2(570f, 60f - 40f);
                }

                currentPositionRotationAxis = positionRotationAxis;
            }
        }

        private void CreatePostionRotationOptionPopups()
        {
            posOptionsPopupGO = UI.AddChildGOImage(parentBoxGO, Color.gray, AnchorPresets.bottomLeft, 75, 105, new Vector2(390, 60));
            rotOptionsPopupGO = UI.AddChildGOImage(parentBoxGO, Color.gray, AnchorPresets.bottomLeft, 75, 45, new Vector2(590, 60));

            var optionButton = UI.CreateUIButton(posOptionsPopupGO, 70, 20, "Zero", 20, 0, -2.5f, AnchorPresets.topMiddle,true);
            optionButton.button.onClick.AddListener(delegate
            {
                ZeroAxisOptionsSelected();
            });
            parentBrowser.ApplyHelp(optionButton, "Set the current axis slider to Zero value");

            optionButton = UI.CreateUIButton(posOptionsPopupGO, 70, 20, "Default", 20, 0, -20 - 2.5f, AnchorPresets.topMiddle, true);
            optionButton.button.onClick.AddListener(delegate
            {
                DefaultAxisOptionsSelected();
            });
            parentBrowser.ApplyHelp(optionButton, "Reset the current axis slider to it's default values");

            optionButton = UI.CreateUIButton(posOptionsPopupGO, 70, 20, "Range x0.1", 20, 0, -40 - 2.5f, AnchorPresets.topMiddle, true);
            optionButton.button.onClick.AddListener(delegate
            {
                DecreaseRangeAxisOptionsSelected();
            });
            parentBrowser.ApplyHelp(optionButton, "Decrease the range of the axis slider to 10% of it's current minimum to maximum range");

            optionButton = UI.CreateUIButton(posOptionsPopupGO, 70, 20, "Range x10", 20, 0, -60 - 2.5f, AnchorPresets.topMiddle, true);
            optionButton.button.onClick.AddListener(delegate
            {
                IncreaseRangeAxisOptionsSelected();
            });
            parentBrowser.ApplyHelp(optionButton, "Increase the min to max range of the axis slider by 10 times.");

            optionButton = UI.CreateUIButton(posOptionsPopupGO, 70, 20, "Re-centre Range", 20, 0, -80 - 2.5f, AnchorPresets.topMiddle, true);
            optionButton.button.onClick.AddListener(delegate
            {
                ResetRangeAxisOptionsSelected();
            });
            parentBrowser.ApplyHelp(optionButton, "Reset the slider min and max values so that the current value is in the centre of the slider");

            optionButton = UI.CreateUIButton(rotOptionsPopupGO, 70, 20, "Zero", 20, 0, -2.5f, AnchorPresets.topMiddle, true);
            optionButton.button.onClick.AddListener(delegate
            {
                ZeroAxisOptionsSelected();
            });
            parentBrowser.ApplyHelp(optionButton, "Set the current axis slider to Zero value");

            optionButton = UI.CreateUIButton(rotOptionsPopupGO, 70, 20, "Default", 20, 0, -20 - 2.5f, AnchorPresets.topMiddle, true);
            optionButton.button.onClick.AddListener(delegate
            {
                DefaultAxisOptionsSelected();
            });
            parentBrowser.ApplyHelp(optionButton, "Reset the current axis slider to it's default values");

            posOptionsPopupGO.SetActive(false);
            rotOptionsPopupGO.SetActive(false);
        }
        private void ZeroAxisOptionsSelected()
        {
            if (currentPositionRotationAxis != -1)
            {
                var jsf = GetSelectedJSF();
                jsf.val = 0f;
                
            }
            posOptionsPopupGO.SetActive(false);
            rotOptionsPopupGO.SetActive(false);
            currentPositionRotationAxis = -1;
        }
        private void DefaultAxisOptionsSelected()
        {
            if (currentPositionRotationAxis != -1)
            {
                var jsf = GetSelectedJSF();
                jsf.val = GetDefaultSelectedJSF().val;

                if (isNewAtomCreation)
                {
                    var jse = GetSelectedRelAbsEnum();
                    jse.val = GetDefaultSelectedRelAbsJSEnum().val;
                }

            }
            posOptionsPopupGO.SetActive(false);
            rotOptionsPopupGO.SetActive(false);
            currentPositionRotationAxis = -1;
        }
        private void DecreaseRangeAxisOptionsSelected()
        {
            if (currentPositionRotationAxis != -1)
            {
                var jsf = GetSelectedJSF();
                float newMin = jsf.min * 0.1f;
                float newMax = jsf.max * 0.1f;
                if (jsf.val < newMin || jsf.val > newMax)
                {
                    float range = newMax - newMin;
                    jsf.min = jsf.val - (range / 2f);
                    jsf.max = jsf.val + (range / 2f);
                }
                else
                {
                    jsf.min = newMin;
                    jsf.max = newMax;
                }
            }
            posOptionsPopupGO.SetActive(false);
            rotOptionsPopupGO.SetActive(false);
            currentPositionRotationAxis = -1;
        }
        private void IncreaseRangeAxisOptionsSelected()
        {
            if (currentPositionRotationAxis != -1)
            {
                var jsf = GetSelectedJSF();
                var range = jsf.max - jsf.min;
                var currentCentre = jsf.min +(range / 2f);

                var newRange = range * 10f;


                jsf.min = currentCentre - (newRange/2f);
                jsf.max = currentCentre + (newRange / 2f);

            }
            posOptionsPopupGO.SetActive(false);
            rotOptionsPopupGO.SetActive(false);
            currentPositionRotationAxis = -1;
        }
        private void ResetRangeAxisOptionsSelected()
        {
            if (currentPositionRotationAxis != -1)
            {
                var jsf = GetSelectedJSF();
                float range = jsf.max - jsf.min;
                jsf.min = jsf.val - (range / 2f);
                jsf.max = jsf.val + (range / 2f);

            }
            posOptionsPopupGO.SetActive(false);
            rotOptionsPopupGO.SetActive(false);
            currentPositionRotationAxis = -1;
        }



        private JSONStorableFloat GetSelectedJSF()
        {
            AtomPositionRotationSync currentTargetAtomPosition = existingAtomPosition;
            if (isNewAtomCreation)
            {
                currentTargetAtomPosition = newAtomPosition;
            }
            if (currentPositionRotationAxis == PositionRotationAxis.xPosition) return currentTargetAtomPosition.positionXJSF;
            if (currentPositionRotationAxis == PositionRotationAxis.yPosition) return currentTargetAtomPosition.positionYJSF;
            if (currentPositionRotationAxis == PositionRotationAxis.zPosition) return currentTargetAtomPosition.positionZJSF;

            if (currentPositionRotationAxis == PositionRotationAxis.xRotation) return currentTargetAtomPosition.rotationXJSF;
            if (currentPositionRotationAxis == PositionRotationAxis.yRotation) return currentTargetAtomPosition.rotationYJSF;
            if (currentPositionRotationAxis == PositionRotationAxis.zRotation) return currentTargetAtomPosition.rotationZJSF;

            return null;

        }
        private JSONStorableEnumStringChooser GetSelectedRelAbsEnum()
        {

            if (currentPositionRotationAxis == PositionRotationAxis.xPosition) return newAtomPosition.posXRelAbsJSE;
            if (currentPositionRotationAxis == PositionRotationAxis.yPosition) return newAtomPosition.posYRelAbsJSE;
            if (currentPositionRotationAxis == PositionRotationAxis.zPosition) return newAtomPosition.posZRelAbsJSE;

            if (currentPositionRotationAxis == PositionRotationAxis.xRotation) return newAtomPosition.rotXRelAbsJSE;
            if (currentPositionRotationAxis == PositionRotationAxis.yRotation) return newAtomPosition.rotYRelAbsJSE;
            if (currentPositionRotationAxis == PositionRotationAxis.zRotation) return newAtomPosition.rotZRelAbsJSE;

            return null;

        }

        private JSONStorableFloat GetDefaultSelectedJSF()
        {
            RelativeAtomPositionRotation defaultNewAtomPositionRotation = Settings.desktopDefaultNewAtomPosRot;
            if (Globals.isVREnabled)
            {
                defaultNewAtomPositionRotation = Settings.vrDefaultNewAtomPosRot;
            }
            if (currentPositionRotationAxis == PositionRotationAxis.xPosition) return defaultNewAtomPositionRotation.positionXJSF;
            if (currentPositionRotationAxis == PositionRotationAxis.yPosition) return defaultNewAtomPositionRotation.positionYJSF;
            if (currentPositionRotationAxis == PositionRotationAxis.zPosition) return defaultNewAtomPositionRotation.positionZJSF;

            if (currentPositionRotationAxis == PositionRotationAxis.xRotation) return defaultNewAtomPositionRotation.rotationXJSF;
            if (currentPositionRotationAxis == PositionRotationAxis.yRotation) return defaultNewAtomPositionRotation.rotationYJSF;
            if (currentPositionRotationAxis == PositionRotationAxis.zRotation) return defaultNewAtomPositionRotation.rotationZJSF;

            return null;

        }

        private JSONStorableEnumStringChooser GetDefaultSelectedRelAbsJSEnum()
        {
            RelativeAtomPositionRotation defaultNewAtomPositionRotation = Settings.desktopDefaultNewAtomPosRot;
            if (Globals.isVREnabled)
            {
                defaultNewAtomPositionRotation = Settings.vrDefaultNewAtomPosRot;
            }
            if (currentPositionRotationAxis == PositionRotationAxis.xPosition) return defaultNewAtomPositionRotation.posXRelAbsJSE;
            if (currentPositionRotationAxis == PositionRotationAxis.yPosition) return defaultNewAtomPositionRotation.posYRelAbsJSE;
            if (currentPositionRotationAxis == PositionRotationAxis.zPosition) return defaultNewAtomPositionRotation.posZRelAbsJSE;

            if (currentPositionRotationAxis == PositionRotationAxis.xRotation) return defaultNewAtomPositionRotation.rotXRelAbsJSE;
            if (currentPositionRotationAxis == PositionRotationAxis.yRotation) return defaultNewAtomPositionRotation.rotYRelAbsJSE;
            if (currentPositionRotationAxis == PositionRotationAxis.zRotation) return defaultNewAtomPositionRotation.rotZRelAbsJSE;

            return null;
        }



        private void RefreshRelativeAtomPopup()
        {
            newAtomPosition.RefreshRelativeAtomChoices();
        }
        private void RefreshRelativeNodePopup()
        {
            newAtomPosition.RefreshRelativeNodeChoices();
        }

        private void TargetAtomTopEnumCallBack(int i)
        {
            newAtomPosition.SetGuideAtomVisible(true);
            RefreshUI();
        }
        private void TargetAtomMainEnumCallBack(string i)
        {
            newAtomPosition.SetGuideAtomVisible(false);
            RefreshUI();
        }

    }


    public class ResourceInfoActionUI : ResourceActionUIBase
    {
        private ResourceSelections resourceSelectionsUI { get { return parentBrowser.resourceSelectionsUI; } }

        protected ResourceVersionGroupEntry primaryRVGE { get { return resourceSelectionsUI.GetFirstSelectedRVGE(filtersResourceType, filtersAtomPresetResourceType); } }

        private BrowserFiltersUI browserFiltersUI { get { return parentBrowser.browserFiltersUI; } }

        private int filtersResourceType { get { return browserFiltersUI.currentResourceType; } }

        private string filtersAtomPresetResourceType { get { return browserFiltersUI.currentAtomPresetResourceType; } }


        private GameObject infoPanelScrollableBoxGO;
        private GameObject infoPaneScrollableViewportContentGO;
        private List<ResourceInfoPanelBase> infoPanelColumns = new List<ResourceInfoPanelBase>();
        private float panelBoxHeight;

        public ResourceInfoActionUI(BrowserUI _parentBrowser, GameObject _parentBoxGO) : base(_parentBrowser, _parentBoxGO)
        {
            panelBoxHeight = parentBrowser.resourceActionBoxHeight - ResourceActionsUI.resourceActionTabsHeight - 4f;

        }

        public override void ClearUI()
        {
            base.ClearUI();
            GameObject.Destroy(infoPanelScrollableBoxGO);
            infoPanelColumns.Clear();
        }

        

        private float hScrollPosition = 0f;
        public override void RefreshUI()
        {
            base.RefreshUI();

            if (resourceSelectionsUI.SelectionCount(filtersResourceType, filtersAtomPresetResourceType) > 0 && primaryRVGE!=null)
            {
                infoPanelScrollableBoxGO = UI.CreateHScrollableContent(parentBoxGO, new Color(0.5f, 0.5f, 0.65f, 0f), AnchorPresets.topLeft,636f , panelBoxHeight, new Vector2(2f, -2f));
                infoPaneScrollableViewportContentGO = infoPanelScrollableBoxGO.transform.Find("scrollableContentViewportGO/hScrollableContentGO").gameObject;

                var scrollRect = infoPanelScrollableBoxGO.GetComponent<ScrollRect>();
                scrollRect.horizontalScrollbar.onValueChanged.AddListener(delegate
                {
                    hScrollPosition = scrollRect.horizontalNormalizedPosition;
                });
                

                infoPanelColumns.Add(new ResourceNameFolderInfoPanel(infoPaneScrollableViewportContentGO, panelBoxHeight,parentBrowser));

                if (!primaryRVGE.localFileResource) infoPanelColumns.Add(new PackageInfoPanel(infoPaneScrollableViewportContentGO, panelBoxHeight, parentBrowser));

                infoPanelColumns.Add(new ThumbnailInfoPanel(infoPaneScrollableViewportContentGO, panelBoxHeight, parentBrowser));
                infoPanelColumns.Add(new NotesInfoPanel(infoPaneScrollableViewportContentGO, panelBoxHeight, parentBrowser));

                scrollRect.horizontalNormalizedPosition = hScrollPosition;
            }

        }

    }

    public class ResourceSaveActionUI : ResourceActionUIBase
    {
        private ResourceSelections resourceSelectionsUI { get { return parentBrowser.resourceSelectionsUI; } }

        protected ResourceVersionGroupEntry primaryRVGE { get { return resourceSelectionsUI.GetFirstSelectedRVGE(filtersResourceType, filtersAtomPresetResourceType); } }

        private BrowserFiltersUI browserFiltersUI { get { return parentBrowser.browserFiltersUI; } }

        private int filtersResourceType { get { return browserFiltersUI.currentResourceType; } }

        private string filtersAtomPresetResourceType { get { return browserFiltersUI.currentAtomPresetResourceType; } }

        private ResourcePrimaryActionUI resourceActionTab;

        private Dictionary<int,string> lastBrowsedSaveFolderByResourceType = new Dictionary<int,string>();

        private Dictionary<string, string> lastBrowsedSaveFolderByAtomResourceType = new Dictionary<string, string>();



        public ResourceSaveActionUI(BrowserUI _parentBrowser, GameObject _parentBoxGO, ResourcePrimaryActionUI _resourceActionTab) : base(_parentBrowser, _parentBoxGO)
        {
            resourceActionTab = _resourceActionTab;
        }

        public override void ClearUI()
        {
            base.ClearUI();

        }

        public override void RefreshUI()
        {
            base.RefreshUI();

            string resourceDescription = "the current Scene.";
            string resourceShortDescription = "Scene";

            float hOffset = edgeWidth;
            if (resourceType != ResourceType.scene)
            {
                resourceActionTab.targetAtomJSMultiEnum.setTopEnumCallbackFunction = TargetAtomTopEnumCallBack;
                resourceActionTab.targetAtomJSMultiEnum.setMainValCallbackFunction = TargetAtomMainEnumCallBack;
                var targetAtomPopup = CreateScrollablePopup(parentBoxGO, resourceActionTab.targetAtomJSMultiEnum, 160f, 25f, "Source Atom", 20, hOffset, -edgeWidth, 500f, true, true, AnchorPresets.topLeft);
                targetAtomPopup.popup.onOpenPopupHandlers += resourceActionTab.RefreshTargetAtomPopupCallback;
                parentBrowser.ApplyHelp(targetAtomPopup, "Select a target atom from the current scene as a source for Preset save.");
                resourceActionTab.RefreshTargetAtomValue();
                
                resourceDescription = "a Preset from the source atom.";
                resourceShortDescription = "Preset";
            }
            hOffset += 170f;

            if (resourceActionTab.targetAtomJSMultiEnum.valType == JSONStorableMultiEnumStringChooser.mainValFlag || resourceType == ResourceType.scene)
            {
                var selectNewResources = CreateToggle(parentBoxGO, Settings.selectNewResourceOnSaveJSB, 160f, 20f, "Select new resource on Save", 20, 0f, -30f, AnchorPresets.topLeft);
                parentBrowser.ApplyHelp(selectNewResources, "Saved resources will be selected in the browser after creation.");

                if (primaryRVGE != null)
                {
                    var duplicateTagDataToggle = CreateToggle(parentBoxGO, Settings.duplicateTagDataOnSaveJSB, 160f, 20f, "Duplicate Tags on Save", 20, 0f, -50f, AnchorPresets.topLeft);
                    parentBrowser.ApplyHelp(duplicateTagDataToggle, "Newly created resources will be automatically set with the same Tags as the currently selected resource");

                }

                var saveButton = CreateUIButton(parentBoxGO, 140f, 30f, "Save new "+ resourceShortDescription + "...", 22, hOffset, -edgeWidth-2f, AnchorPresets.topLeft);
                saveButton.button.onClick.AddListener(delegate
                {
                    SaveNew(true);
                });
                saveButton.buttonColor = new Color(0f, 0.6f, 0);
                saveButton.textColor = Color.white;
                parentBrowser.ApplyHelp(saveButton, "Opens a dialogue to select a filename and folder which will save "+resourceDescription);

                if (resourceType!=ResourceType.scene)
                {
                    var saveButtonNoScreenshot = CreateUIButton(parentBoxGO, 140f, 30f, "Save new "+ resourceShortDescription + "...\n(no screenshot)", 22, hOffset, -edgeWidth - 37f, AnchorPresets.topLeft);
                    saveButtonNoScreenshot.buttonColor = new Color(0f, 0.6f, 0);
                    saveButtonNoScreenshot.textColor = Color.white;
                    saveButtonNoScreenshot.button.onClick.AddListener(delegate
                    {
                        SaveNew(false);
                    });
                    parentBrowser.ApplyHelp(saveButtonNoScreenshot, "Opens a dialogue to select a filename and folder which will save " + resourceDescription+" No screenshot will be captured as a resource thumbnail image.");
                }
                

                hOffset += 150f;

                var saveVersionButton = CreateUIButton(parentBoxGO, 140f, 30f, "Save new "+ resourceShortDescription + " variant", 22, hOffset, -edgeWidth-2f, AnchorPresets.topLeft);
                saveVersionButton.buttonColor = new Color(0f, 0.6f, 0);
                saveVersionButton.textColor = Color.white;
                saveVersionButton.button.onClick.AddListener(delegate
                {
                    SaveNewVariant(true);
                });
                parentBrowser.ApplyHelp(saveVersionButton, "Instantly creates a new variant of "+resourceDescription+ " The new file will have a unique name derived from the currently selected resource and be stored in the same folder.");

                if (resourceType != ResourceType.scene)
                {
                    var saveVersionButtonNoScreenshot = CreateUIButton(parentBoxGO, 140f, 30f, "Save new "+ resourceShortDescription + " variant\n(no screenshot)", 22, hOffset, -edgeWidth - 37f, AnchorPresets.topLeft);
                    saveVersionButtonNoScreenshot.buttonColor = new Color(0f, 0.6f, 0);
                    saveVersionButtonNoScreenshot.textColor = Color.white;
                    saveVersionButtonNoScreenshot.button.onClick.AddListener(delegate
                    {
                        SaveNewVariant(false);
                    });
                    parentBrowser.ApplyHelp(saveVersionButtonNoScreenshot, "Instantly creates a new variant of " + resourceDescription + " The new file will have a unique name derived from the currently selected resource and be stored in the same folder. No screenshot will be captured, but a duplicate of the current resources thumbnail will be used.");
                }
                    
                hOffset += 150f;

                if (primaryRVGE != null && primaryRVGE.localFileResource)
                {
                    var overwriteCurrentSelectionButton = CreateUIButton(parentBoxGO, 140f, 30f, "Overwrite "+ resourceShortDescription + "...", 22, hOffset, -edgeWidth - 2f, AnchorPresets.topLeft);
                    overwriteCurrentSelectionButton.buttonColor = new Color(1f, 0, 0);
                    overwriteCurrentSelectionButton.textColor = Color.white;
                    overwriteCurrentSelectionButton.button.onClick.AddListener(delegate
                    {
                        Overwrite(true);
                    });
                    parentBrowser.ApplyHelp(overwriteCurrentSelectionButton, "Overwrites the selected resource with "+resourceDescription);

                    if (resourceType != ResourceType.scene)
                    {
                        var overwriteCurrentNoScreenshotButton = CreateUIButton(parentBoxGO, 140f, 30f, "Overwrite "+ resourceShortDescription + "...\n(no screenshot)", 22, hOffset, -edgeWidth - 37f, AnchorPresets.topLeft);
                        overwriteCurrentNoScreenshotButton.buttonColor = new Color(1f, 0, 0);
                        overwriteCurrentNoScreenshotButton.textColor = Color.white;
                        overwriteCurrentNoScreenshotButton.button.onClick.AddListener(delegate
                        {
                            Overwrite(false);
                        });
                        parentBrowser.ApplyHelp(overwriteCurrentNoScreenshotButton, "Overwrites the selected resource with " + resourceDescription+" No new screenshot will be captured as a resource thumbnail.");
                    }                      
                }
            }

        }
        private void TargetAtomTopEnumCallBack(int i)
        {
            if (parentBrowser.resourceActionsUI.getCurrentActionUITab == this) RefreshUI();
        }
        private void TargetAtomMainEnumCallBack(string i)
        {
            if (parentBrowser.resourceActionsUI.getCurrentActionUITab == this) RefreshUI();
        }

        private bool takeScreenShot = false;
        private void SaveNew(bool screenshot)
        {
            takeScreenShot = screenshot;
            string prefixRemoval = null;
            string fileExtFilter = ResourceType.GetFileExtensionFilter(resourceType);
            string folderName = GetLastBrowsedFolder();
            if (resourceCategory==ResourceCategory.preset) prefixRemoval = "Preset_";
            SuperController.singleton.GetMediaPathDialog(SaveNewFileSelected, fileExtFilter, folderName, false, true, false, prefixRemoval, true);

            uFileBrowser.FileBrowser browser = SuperController.singleton.mediaFileBrowserUI;
            browser.SetTextEntry(true);
            browser.fileEntryField.text = String.Format("{0}", ((int)(DateTime.UtcNow - new DateTime(1970, 1, 1)).TotalSeconds).ToString());
            browser.ActivateFileNameField();

        }

        private void SaveNewFileSelected(string fileNamePath)
        {            
            if (fileNamePath == null || fileNamePath == "") return;

            var rootFolderNames = Const.GetResourceRootFolderNames(resourceType, atomPresetResourceType);

            bool inRootFolder = false;
            foreach (var rootFolder in rootFolderNames)
            {
                if (fileNamePath.StartsWith(rootFolder))
                {
                    inRootFolder = true;
                    break;
                }
            }

            if (!inRootFolder)
            {
                SuperController.singleton.Alert("Resource file location must be in a recognised folder for this resource type such as " + rootFolderNames.First(),null);
                return;
            }

            StoreLastBrowsedFolder(fileNamePath);

            fileNamePath = fileNamePath + "." + ResourceType.GetPrimaryFileExtension(resourceType);
            if (resourceCategory == ResourceCategory.preset) fileNamePath = FileManagerSecure.GetDirectoryName(fileNamePath)+"\\Preset_"+FileManagerSecure.GetFileName(fileNamePath);

            if (FileManagerSecure.FileExists(fileNamePath))
            {
                SuperController.singleton.Alert("Resource "+fileNamePath+" already exists. Please confirm that you wish to overwrite the file.", delegate
                {
                    
                    SaveFinal(fileNamePath, takeScreenShot);
                }, delegate
                {
                    // DO NOTHING
                });
            }
            else SaveFinal(fileNamePath, takeScreenShot);

        }

        private void StoreLastBrowsedFolder (string path)
        {
            if (path== null || path=="") return;
            if (resourceType != ResourceType.presetAtom)
            {
                lastBrowsedSaveFolderByResourceType.Add(resourceType, FileManagerSecure.GetDirectoryName(path));
            }
            else lastBrowsedSaveFolderByAtomResourceType.Add(atomPresetResourceType, FileManagerSecure.GetDirectoryName(path));
        }

        private string GetLastBrowsedFolder()
        {
            string folder = Const.GetResourceRootFolderNames(resourceType, atomPresetResourceType).First();
            if (resourceType != ResourceType.presetAtom)
            {
                if (lastBrowsedSaveFolderByResourceType.ContainsKey(resourceType)) folder = lastBrowsedSaveFolderByResourceType[resourceType];
            }
            else if (lastBrowsedSaveFolderByAtomResourceType.ContainsKey(atomPresetResourceType)) folder = lastBrowsedSaveFolderByAtomResourceType[atomPresetResourceType];

            return folder;
        }


        private void SaveNewVariant(bool screenshot)
        {
            var currentSelection = parentBrowser.resourceSelectionsUI.GetFirstSelectedRVGE(resourceType, atomPresetResourceType);

            int i = 1;
            while (FileManagerSecure.FileExists(currentSelection.folderPath+"\\"+ currentSelection.resourceFileNameNoExt + "(" + i.ToString() + ")." + currentSelection.resourceFileNameExtension))
            {
                i++;
            }
            string path = currentSelection.folderPath + "\\" + currentSelection.resourceFileNameNoExt + "(" + i.ToString() + ")." + currentSelection.resourceFileNameExtension;

            SaveFinal(path, screenshot);

            if (!screenshot && ((Globals.customFolderAccess && resourceCategory==ResourceCategory.preset) || (Globals.scenesFolderAccess && resourceType==ResourceType.scene))&& FileManagerSecure.FileExists(currentSelection.resourceFolderPathFileNameNoExt+".jpg"))
            {
                var rvge = ResourceManifest.GetRVGE(path) as LocalRVGE;

                FileManagerSecure.CopyFile(currentSelection.resourceFolderPathFileNameNoExt + ".jpg", Const.SymLinksFolder + "\\" + rvge.resourceFolderPathFileNameNoExt+ ".jpg");
                RefreshRVGEThumbnail(rvge);
            }
        }

        private void Overwrite(bool screenshot)
        {
            var currentSelection = parentBrowser.resourceSelectionsUI.GetFirstSelectedRVGE(resourceType, atomPresetResourceType);

            var fileNamePath = currentSelection.resourceFullPathReference;

            if (FileManagerSecure.FileExists(fileNamePath))
            {
                SuperController.singleton.Alert("Resource " + fileNamePath + " already exists. Please confirm that you wish to overwrite the file.", delegate
                {
                    SaveFinal(fileNamePath, screenshot);
                }, delegate
                {
                    // DO NOTHING
                });
            }
            else SaveFinal(fileNamePath, screenshot);
        }

        private void SaveFinal(string path, bool screenshot)
        {
            if (resourceType == ResourceType.scene)
            {
                if (Globals.scenesFolderAccess) path = Const.SymLinksFolder + "\\" + path;
                SuperController.singleton.Save(path);
            }
            else if (resourceCategory == ResourceCategory.preset)
            {
                if (resourceActionTab.targetAtomJSMultiEnum.valType == JSONStorableMultiEnumStringChooser.mainValFlag)
                {
                    Atom sourceAtom = SuperController.singleton.GetAtomByUid(resourceActionTab.targetAtomJSMultiEnum.mainVal);
                    if (sourceAtom != null)
                    {
                        JSONStorable presetJS = null;
                        if (resourceType == ResourceType.presetAppearance) presetJS = sourceAtom.GetStorableByID("AppearancePresets");
                        else if (PC.patreonContentEnabled) presetJS = PC.GetPresetStorable(sourceAtom, resourceType);
                        if (presetJS != null)
                        {
                            JSONStorableBool loadOnSelectJSB = presetJS.GetBoolJSONParam("loadPresetOnSelect");

                            bool loadOnSelectPreState = loadOnSelectJSB.val;
                            loadOnSelectJSB.val = false;

                            JSONStorableUrl presetPathJSON = presetJS.GetUrlJSONParam("presetBrowsePath");
                            presetPathJSON.val = SuperController.singleton.NormalizePath(path);
                            if (screenshot)
                            {
                                presetJS.CallAction("StorePresetWithScreenshot");
                                
                            }
                            else presetJS.CallAction("StorePreset");

                            loadOnSelectJSB.val = loadOnSelectPreState;
                        }
                    }
                }
            }
            else return;

            if (FileManagerSecure.FileExists(path))
            {
                LocalRVGE rvge = ResourceManifest.GetRVGE(path) as LocalRVGE;

                bool preExists = rvge != null;

                if (!preExists) rvge = ResourceManifest.RegisterLocalResourceOfType(path, resourceType, atomPresetResourceType, DateTime.Now, false);
                else
                {
                    rvge.isNew = true;
                    rvge.UpdateLocalFileCreateUpdateDates();
                }

                if (screenshot) parentBrowser.SetScreenShotCallback(RefreshRVGEThumbnail, rvge);

                if (Settings.duplicateTagDataOnSaveJSB.val && !preExists)
                {
                    var currentSelection = parentBrowser.resourceSelectionsUI.GetFirstSelectedRVGE(resourceType, atomPresetResourceType);

                    if (currentSelection != null)
                    {
                        foreach (Tag tag in TagManifest.GetRVGETags(currentSelection))
                        {
                            if (tag.tagCategory != TagCategory.combi)
                            {
                                rvge.SetTag(tag, true);
                            }

                        }
                    }

                    if (currentSelection.isFavorite) rvge.isFavorite = true;

                }

                parentBrowser.RefreshFilterResults();

                if (Settings.selectNewResourceOnSaveJSB.val)
                {
                    parentBrowser.resourceSelectionsUI.SingleSelectRVGE(rvge, parentBrowser.resourceBrowserUI.GetUIDButton(rvge));
                }
            }
            else SuperController.LogError("BA.ResourceSaveActionUI.SaveFinal: Failed to create new save "+path);
        }

        public void RefreshRVGEThumbnail(ResourceVersionGroupEntry rvge)
        {
            parentBrowser.resourceBrowserUI.RefreshRVGEThumbnail(rvge);
        }
    }



    public abstract class ResourceInfoPanelBase
    {
        protected GameObject currentPanelColumnGO;
        protected BrowserUI parentBrowser;
        protected ResourceVersionGroupEntry primaryRVGE { get { return resourceSelectionsUI.GetFirstSelectedRVGE(resourceType, atomPresetResourceType); } }

        protected ResourceSelections resourceSelectionsUI { get { return parentBrowser.resourceSelectionsUI; } }

        private int resourceType { get { return parentBrowser.browserFiltersUI.currentResourceType; } }
        private string atomPresetResourceType { get { return parentBrowser.browserFiltersUI.currentAtomPresetResourceType; } }

        public ResourceInfoPanelBase(GameObject parentViewportGO, float panelBoxHeight, BrowserUI parentBrowser )
        {
            currentPanelColumnGO = UI.AddChildGOImage(parentViewportGO, new Color(0f, 0f, 0f, 0f), AnchorPresets.vStretchLeft, 250f, panelBoxHeight, Vector2.zero);
            this.parentBrowser = parentBrowser;

        }
    }

    public class ResourceNameFolderInfoPanel: ResourceInfoPanelBase
    {
        private UIDynamicButton renameResourceButton;
        private GameObject fileNameTextGO;
        private UIDynamicTextField fileNameText;
        public ResourceNameFolderInfoPanel(GameObject parentViewportGO, float panelBoxHeight, BrowserUI parentBrowser) : base (parentViewportGO, panelBoxHeight, parentBrowser)
        {
            if (primaryRVGE != null)
            {
                float resourceFilenameBoxWidth = 245f;
                if (primaryRVGE.localFileResource && primaryRVGE.isRenameable)
                {
                    resourceFilenameBoxWidth = 195f;
                    renameResourceButton = UI.CreateUIButton(currentPanelColumnGO, 50f, 20f, "Rename", 20, 0, 0, AnchorPresets.topRight);
                    renameResourceButton.button.onClick.AddListener(delegate
                    {
                        renameResourceButton.gameObject.SetActive(false);
                        fileNameTextGO.SetActive(false);
                        fileNameText.gameObject.SetActive(true);
                        var inputComponent = fileNameText.gameObject.GetComponent<InputField>();
                        inputComponent.ActivateInputField();
                    });
                    parentBrowser.ApplyHelp(renameResourceButton, "Rename the resource file. This will change the actual filename and will be applicable outside of BrowserAssist, potentially breaking any dependencies on this file.");
                }
                fileNameTextGO = UI.CreateBasicTextBox(currentPanelColumnGO, primaryRVGE.resourceFileNameNoExtNoPrefix, resourceFilenameBoxWidth, 20f, 20, 0, 0, AnchorPresets.topLeft, backgoundColor: BrowserUI.lightBlueGrey);
                parentBrowser.ApplyHelp(fileNameTextGO, "Default Resource name is " + primaryRVGE.resourceFileNameNoExtNoPrefix + ". This is based on the actual filename of the resource.");

                var fileNameJSS = new JSONStorableString("fileNameJSS",primaryRVGE.resourceFileNameNoExtNoPrefix, FileNameUpdateCallback);
                fileNameText = UI.CreateTextBox(currentPanelColumnGO, fileNameJSS, 245f, 20f, 20, 0f, 0f, AnchorPresets.topLeft, true, InputField.LineType.SingleLine, "", 0, 0, null, TextAnchor.MiddleLeft, FileNameUpdateEndCallback);
                parentBrowser.ApplyHelp(fileNameText, "The resource Alias name is an alternative name that can be specified by the user and will override the default resource name. It is only used within BrowserAssist.");

                fileNameText.gameObject.SetActive(false);

                var aliasText = UI.CreateTextBox(currentPanelColumnGO, primaryRVGE.resourceNameAliasJSS, 245f, 20f, 20, 0f, -20f, AnchorPresets.topLeft, true, InputField.LineType.SingleLine, "Enter Alias for Resource Name", 0, 0,null,TextAnchor.MiddleLeft);
                parentBrowser.ApplyHelp(aliasText, "The resource Alias name is an alternative name that can be specified by the user and will override the default resource name. It is only used within BrowserAssist.");
                var clearAliasButton = UI.CreateUIButton(currentPanelColumnGO, 20f, 20f, "X", 20, -5f, -20f, AnchorPresets.topRight);
                parentBrowser.ApplyHelp(clearAliasButton, "Clear the alias name");
                clearAliasButton.button.onClick.AddListener(delegate { primaryRVGE.resourceNameAliasJSS.val = ""; });

                GameObject folderTextBoxGO;
                if (primaryRVGE.localFileResource)
                {
                    folderTextBoxGO=UI.CreateBasicTextBox(currentPanelColumnGO, new JSONStorableString("folder", primaryRVGE.folderPath), 215f, 20f, 20, 0, -40f, AnchorPresets.topLeft, true, backgoundColor: BrowserUI.lightBlueGrey);
                    var filterButton = UI.CreateUIButton(currentPanelColumnGO, 30f, 20f, "Filter", 20, 215f, -40f, AnchorPresets.topLeft, false);
                    parentBrowser.ApplyHelp(filterButton, "Apply this resource's folder as a Filter in the resource browser.");
                    filterButton.button.onClick.AddListener(delegate
                    {
                        parentBrowser.browserFiltersUI.FolderFilterSelected(primaryRVGE.containingLocalFolder);
                        parentBrowser.browserFiltersUI.creatorJSMultiEnum.valTopEnum = CreatorFilterChoices.localOnly;
                        
                    });

                }
                else folderTextBoxGO = UI.CreateBasicTextBox(currentPanelColumnGO, new JSONStorableString("folder", primaryRVGE.folderPath), 245f, 20f, 20, 0, -40f, AnchorPresets.topLeft,true, backgoundColor: BrowserUI.lightBlueGrey);
                parentBrowser.ApplyHelp(folderTextBoxGO, primaryRVGE.folderPath+"\nFolder that the resource is located within");

            }
            
        }
        private void FileNameUpdateEndCallback (string fileName)
        {
            renameResourceButton.gameObject.SetActive(true);
            fileNameTextGO.SetActive(true);
            fileNameText.gameObject.SetActive(false);
        }
        private void FileNameUpdateCallback(string newFileName)
        {
            renameResourceButton.gameObject.SetActive(true);
            fileNameTextGO.SetActive(true);
            fileNameText.gameObject.SetActive(false);

            char[] illegalChars = { '<', '>', ':', '"', '/','\\','|','?','*' };

            if (newFileName.IndexOfAny(illegalChars)==-1)
            {
                string newFileNameWithNoExt = newFileName;
                if (primaryRVGE.resourceCategory == ResourceCategory.preset) newFileNameWithNoExt = "Preset_" + newFileNameWithNoExt;
                string newFullPathFileNameNoExt = primaryRVGE.folderPath + "\\" + newFileNameWithNoExt;
                string newFullPathFileNameWithExt = newFullPathFileNameNoExt + "." + primaryRVGE.resourceFileNameExtension;
                if (!FileManagerSecure.FileExists(newFullPathFileNameWithExt))
                {
                    string sourceFolderPathFileName = Const.SymLinksFolder + "\\" + primaryRVGE.resourceFolderPathFileName;
                    string sourceFolderPathFileNameNoExt = Const.SymLinksFolder + "\\" + primaryRVGE.resourceFolderPathFileNameNoExt;
                    string targetFolderPathFileNameNoExt = Const.SymLinksFolder + "\\" + newFullPathFileNameNoExt;
                    string targetFolderPathFileName = Const.SymLinksFolder + "\\" + newFullPathFileNameWithExt;

                    try
                    {
                        FileManagerSecure.MoveFile(sourceFolderPathFileName, targetFolderPathFileName);
                    }
                    catch
                    {
                        SuperController.LogMessage("Move Selected Resources failed for '" + sourceFolderPathFileName + "'. Do you perhaps have the resource file open in a text editor?");
                        return;
                    }

                    try
                    {
                        if (FileManagerSecure.FileExists(sourceFolderPathFileName + ".fav"))
                        {
                            FileManagerSecure.MoveFile(sourceFolderPathFileName + ".fav", targetFolderPathFileName + ".fav");
                        }

                    }
                    catch
                    {
                        SuperController.LogMessage("Move Selected Resources Favourite Tag failed for '\"+ sourceFolderPathFileName+\"'.");
                    }

                    try
                    {
                        if (FileManagerSecure.FileExists(sourceFolderPathFileName + ".hide"))
                        {
                            FileManagerSecure.MoveFile(sourceFolderPathFileName + ".hide", targetFolderPathFileName + ".hide");
                        }

                    }
                    catch
                    {
                        SuperController.LogMessage("Move Selected Resources Hidden Tag failed for '\"+ sourceFolderPathFileName+\"'.");
                    }


                    try
                    {
                        if (FileManagerSecure.FileExists(sourceFolderPathFileNameNoExt + ".jpg"))
                        {
                            FileManagerSecure.MoveFile(sourceFolderPathFileNameNoExt + ".jpg", targetFolderPathFileNameNoExt + ".jpg");
                        }

                    }
                    catch
                    {
                        SuperController.LogMessage("Move Selected Resources Thumbnail image failed for '\"+ sourceFolderPathFileName+\"'.");
                    }

                    ResourceManifest.RenameRVGEFilename(primaryRVGE as LocalRVGE, newFullPathFileNameWithExt);
                    parentBrowser.resourceBrowserUI.UpdateDisplayName(primaryRVGE);
                    parentBrowser.resourceSelectionsUI.UpdateDisplayName(primaryRVGE);
                    parentBrowser.resourceActionsUI.RefreshUI();
                }
                else
                {
                    SuperController.LogMessage("BrowserAssist: The entered filename already exists in the folder. Chose an alternative filename.");
                }

            }
            else
            {
                SuperController.LogMessage("BrowserAssist: The entered filename contains illegal characters not allowed by Windows Operating System");
            }

        }
    }

    public class PackageInfoPanel : ResourceInfoPanelBase
    {
        public PackageInfoPanel(GameObject parentViewportGO, float panelBoxHeight, BrowserUI parentBrowser) : base(parentViewportGO, panelBoxHeight, parentBrowser)
        {
            if (primaryRVGE != null && !primaryRVGE.localFileResource)
            {
                PackagedRVGE packagedRVGE = primaryRVGE as PackagedRVGE;

                UI.CreateBasicTextBox(currentPanelColumnGO, "Creator:", 50f, 20f, 20, 0, 0, AnchorPresets.topLeft, backgoundColor: new Color(0,0,0,0));
                var textBox = UI.CreateBasicTextBox(currentPanelColumnGO,  packagedRVGE.creatorName, 165f, 20f, 20, 50f, 0, AnchorPresets.topLeft, backgoundColor: BrowserUI.lightBlueGrey);
                parentBrowser.ApplyHelp(textBox, "Name of the Creator who packaged the resource");

                var creatorFilterButton = UI.CreateUIButton(currentPanelColumnGO, 30f, 20f, "Filter", 20, 215f, 0f, AnchorPresets.topLeft, false);
                creatorFilterButton.button.onClick.AddListener(delegate { parentBrowser.browserFiltersUI.creatorJSMultiEnum.mainVal = primaryRVGE.creatorName; });
                parentBrowser.ApplyHelp(creatorFilterButton, "Update the filters so that only resources from this Creator are selectable");

                UI.CreateBasicTextBox(currentPanelColumnGO, "Package:", 50f, 20f, 20, 0, -20f, AnchorPresets.topLeft, backgoundColor: new Color(0, 0, 0, 0));
                textBox = UI.CreateBasicTextBox(currentPanelColumnGO, packagedRVGE.packageVersionUID, 165f, 20f, 20, 50f,-20f, AnchorPresets.topLeft, backgoundColor: BrowserUI.lightBlueGrey);
                parentBrowser.ApplyHelp(textBox, "Full package name that contains the resource");

                var packageRescanButton = UI.CreateUIButton(currentPanelColumnGO, 30f, 20f, "Rescan", 16, 215f, -20f, AnchorPresets.topLeft, false);
                packageRescanButton.button.onClick.AddListener(delegate { VARPackageManifest.RescanVARResources(packagedRVGE.vpvge, packagedRVGE.highestVARVersion, new HashSet<int>());
                    parentBrowser.RefreshFilterResults(true);
                });
                parentBrowser.ApplyHelp(packageRescanButton, "Rescan the contents of the containing VAR Package.");



                UI.CreateBasicTextBox(currentPanelColumnGO, "License:", 50f, 20f, 20, 0, -40f, AnchorPresets.topLeft, backgoundColor: new Color(0, 0, 0, 0));
                textBox = UI.CreateBasicTextBox(currentPanelColumnGO,  EnumManifest.GetDisplayVal(LicenseType.enumManifestName, packagedRVGE.varLicenseType), 165f, 20f, 20, 50, -40f, AnchorPresets.topLeft, backgoundColor: BrowserUI.lightBlueGrey);
                parentBrowser.ApplyHelp(textBox, "License Type that applies to package containing the resource");

                var licenseFilterButton = UI.CreateUIButton(currentPanelColumnGO, 30f, 20f, "Filter", 20, 215f, -40f, AnchorPresets.topLeft, false);
                licenseFilterButton.button.onClick.AddListener(delegate { parentBrowser.browserFiltersUI.licenseFilterJSEnum.val = packagedRVGE.varLicenseType; });
                parentBrowser.ApplyHelp(licenseFilterButton, "Update the filters so that only resources with this License are selectable");

            }

        }
    }

    public class ThumbnailInfoPanel : ResourceInfoPanelBase
    {
        public ThumbnailInfoPanel(GameObject parentViewportGO, float panelBoxHeight, BrowserUI parentBrowser) : base(parentViewportGO, panelBoxHeight, parentBrowser)
        {
            if (primaryRVGE != null)
            {
                var imageGO = UI.AddChildGOImage(currentPanelColumnGO,Color.white,AnchorPresets.topLeft,60f,60f,new Vector2(0,0),primaryRVGE.resourceSprite);
                if (FileManagerSecure.FileExists(FileManagerSecure.NormalizePath(primaryRVGE.resourceThumbnailReference))) parentBrowser.ApplyHelp(imageGO, primaryRVGE.resourceThumbnailReference);
                else parentBrowser.ApplyHelp(imageGO, "No thumbnail image exists for this resource.");

                if (!primaryRVGE.localFileResource || primaryRVGE.isRenameable)
                {
                    var button = UI.CreateUIButton(currentPanelColumnGO, 185f, 20f, "Take screenshot for Thumbnail...", 20, 60f, 0f, AnchorPresets.topLeft);
                    button.button.onClick.AddListener(delegate {
                        if (primaryRVGE.localFileResource)
                        {
                            SuperController.singleton.DoSaveScreenshot(Const.SymLinksFolder + "\\"+primaryRVGE.resourceFolderPathFileNameNoExt+".jpg",ResourceScreenShotCallback);
                            
                        }
                        else
                        {
                            PackagedRVGE rvge = primaryRVGE as PackagedRVGE;
                            
                            FileManagerSecure.CreateDirectory(rvge.customThumbnailFolder);

                            SuperController.singleton.DoSaveScreenshot(rvge.customThumbnailFile, ResourceScreenShotCallback);

                        }
                        
                    });
                    if (primaryRVGE.localFileResource)
                    {
                        parentBrowser.ApplyHelp(button, "Take a VAM screenshot to replace the Thumbnail image for this resource. The thumbnail will overwrite the original thumbnail and will be used by the VAM native File Browser.",true);
                    }
                    else
                    {
                        parentBrowser.ApplyHelp(button, "Take a VAM screenshot to use as a custom Thumbnail image for this resource. The custom thumbnail will not overwrite the original thumbnail and will only be used in BrowserAssist.");
                    }

                    button = UI.CreateUIButton(currentPanelColumnGO, 185f, 20f, "Select thumbnail image from disk...", 20, 60f, -20f, AnchorPresets.topLeft);
                    button.button.onClick.AddListener(delegate
                    {
                        SuperController.singleton.GetMediaPathDialog(ThumbnailFileCallback, "jpg","",false,true,showInstallFolderInDirectoryList:true);
                    });

                    if (primaryRVGE.localFileResource)
                    {
                        parentBrowser.ApplyHelp(button, "Select an image file from disk storage to replace the Thumbnail image for this resource. The thumbnail will overwrite the original thumbnail and will be used by the VAM native File Browser.", true);
                    }
                    else
                    {
                        parentBrowser.ApplyHelp(button, "Select an image from file system to use as a custom Thumbnail image for this resource. The custom thumbnail will not overwrite the original thumbnail and will only be used in BrowserAssist.");
                    }


                    if (!primaryRVGE.localFileResource)
                    {
                        PackagedRVGE rvge = primaryRVGE as PackagedRVGE;
                        if (FileManagerSecure.FileExists(rvge.customThumbnailFile))
                        {
                            button = UI.CreateUIButton(currentPanelColumnGO, 185f, 20f, "Reset thumbnail to default", 20, 60f, -40f, AnchorPresets.topLeft);
                            button.buttonColor = Color.red;
                            button.buttonText.color = Color.white;
                            parentBrowser.ApplyHelp(button, "Clear the custom Thumbnail image and revert to using the default resource thumbnail image supplied by the VAR creator. This will immediately and permanently delete the custom thumbnail.", true);
                            button.button.onClick.AddListener(delegate
                            {
                                FileManagerSecure.DeleteFile(rvge.customThumbnailFile);
                                primaryRVGE.ReloadSprite();
                            }
                            );
                        }

                    }
                    
                }


            }

        }

        private void ThumbnailFileCallback(string fileName)
        {
            if (fileName != "")
            {
                if (primaryRVGE.localFileResource)
                {
                    string targetThumbnail = Const.SymLinksFolder + "\\" + primaryRVGE.resourceFolderPathFileNameNoExt + ".jpg";
                    FileManagerSecure.DeleteFile(targetThumbnail);
                    FileManagerSecure.CopyFile(fileName, targetThumbnail);
                }
                else
                {
                    PackagedRVGE rvge = primaryRVGE as PackagedRVGE;

                    FileManagerSecure.CreateDirectory(rvge.customThumbnailFolder);
                    FileManagerSecure.DeleteFile(rvge.customThumbnailFile);
                    FileManagerSecure.CopyFile(fileName, rvge.customThumbnailFile);
                }
                primaryRVGE.ReloadSprite();
            }

        }
        private void ResourceScreenShotCallback(string text)
        {
          
            primaryRVGE.ReloadSprite();
            parentBrowser.SetActive();
        }
    }

    public class NotesInfoPanel : ResourceInfoPanelBase
    {
        public NotesInfoPanel(GameObject parentViewportGO, float panelBoxHeight, BrowserUI parentBrowser) : base(parentViewportGO, panelBoxHeight, parentBrowser)
        {
            if (primaryRVGE != null)
            {
                
                UI.CreateTextBox(currentPanelColumnGO, primaryRVGE.notesJSS, 245f, 60f, 20, 0f, 0f, AnchorPresets.topLeft, true, InputField.LineType.MultiLineSubmit, "Enter any notes related to this resource", 0, 10f, null, TextAnchor.UpperLeft);


            }

        }
    }



}
